import pandas as pd
from darts.models import (
    CatBoostModel as CBT,
    LightGBMModel as LGB,
    XGBModel as XGB,
    RandomForest as RF
)
from spinesUtils.asserts import generate_function_kwargs, ParameterTypeAssert
from PipelineTS.base.base import GBDTModelMixin, IntervalEstimationMixin
from PipelineTS.base.darts_base import DartsForecastMixin
from PipelineTS.utils import check_time_col_is_timestamp


class CatBoostModel(DartsForecastMixin, GBDTModelMixin, IntervalEstimationMixin):
    @ParameterTypeAssert({
        'time_col': str,
        'target_col': str,
        'lags': int,
        'lags_past_covariates': (None, int),
        'lags_future_covariates': (None, int),
        'add_encoders': (dict, None),
        'quantile': (None, float),
        'random_state': (None, int),
        'multi_models': bool,
        'use_static_covariates': bool,
        'verbose': bool
    }, 'CatBoostModel')
    def __init__(
            self,
            time_col,
            target_col,
            lags=1,
            lags_past_covariates=None,
            lags_future_covariates=None,
            add_encoders=None,
            quantile=0.9,
            random_state=None,
            multi_models=True,
            use_static_covariates=True,
            verbose=False,
            **darts_catboost_model_configs
    ):
        """
        Initialize the CatBoostModel.

        Parameters
        ----------
        time_col : str
            The name of the time column in the input DataFrame.

        target_col : str
            The name of the target column in the input DataFrame.

        lags : int
            The number of lagged time steps to consider. Default is 1.

        lags_past_covariates : int or None, optional
            The number of lagged time steps for past covariates. Default is None.

        lags_future_covariates : int or None, optional
            The number of lagged time steps for future covariates. Default is None.

        add_encoders : dict or None, optional
            Additional encoders for categorical variables. Default is None.

        quantile : float or None, optional
            The quantile level for prediction intervals. Default is 0.9 (90%).

        random_state : int or None, optional
            The random seed for reproducibility. Default is None.

        multi_models : bool, optional
            Whether to use multiple models. Default is True.

        use_static_covariates : bool, optional
            Whether to use static covariates. Default is True.

        verbose : bool, optional
            Whether to print verbose output. Default is False.

        **darts_catboost_model_configs
            Additional configurations specific to the CatBoost model.
        """
        super().__init__(time_col=time_col, target_col=target_col)

        # Generate model configurations
        self.all_configs['model_configs'] = generate_function_kwargs(
            CBT,
            lags=lags + 1,
            lags_past_covariates=lags_past_covariates,
            lags_future_covariates=lags_future_covariates,
            output_chunk_length=lags,
            add_encoders=add_encoders,
            random_state=random_state,
            multi_models=multi_models,
            use_static_covariates=use_static_covariates,
            verbose=verbose,
            **darts_catboost_model_configs
        )

        # Define and initialize the model
        self.model = self._define_model()

        # Update configurations
        self.all_configs.update(
            {
                'lags': lags,
                'quantile': quantile,
                'time_col': time_col,
                'target_col': target_col,
                'quantile_error': 0
            }
        )

    def _define_model(self):
        """Define the CatBoost model."""
        return CBT(**self.all_configs['model_configs'])


class LightGBMModel(DartsForecastMixin, GBDTModelMixin, IntervalEstimationMixin):
    @ParameterTypeAssert({
        'time_col': str,
        'target_col': str,
        'lags': int,
        'lags_past_covariates': (None, int),
        'lags_future_covariates': (None, int),
        'add_encoders': (dict, None),
        'quantile': (None, float),
        'random_state': (None, int),
        'multi_models': bool,
        'use_static_covariates': bool,
        'categorical_past_covariates': (str, list, None),
        'categorical_future_covariates': (str, list, None),
        'categorical_static_covariates': (str, list, None),
        'verbose': int
    }, 'LightGBMModel')
    def __init__(
            self,
            time_col,
            target_col,
            lags=1,
            lags_past_covariates=None,
            lags_future_covariates=None,
            add_encoders=None,
            quantile=0.9,
            random_state=None,
            multi_models=True,
            use_static_covariates=True,
            categorical_past_covariates=None,
            categorical_future_covariates=None,
            categorical_static_covariates=None,
            verbose=-1,
            linear_tree=True,
            **darts_lightgbm_model_configs
    ):
        """
        Initialize the LightGBMModel.

        Parameters
        ----------
        time_col : str
            The name of the time column in the input DataFrame.

        target_col : str
            The name of the target column in the input DataFrame.

        lags : int
            The number of lagged time steps to consider. Default is 1.

        lags_past_covariates : int or None, optional
            The number of lagged time steps for past covariates. Default is None.

        lags_future_covariates : int or None, optional
            The number of lagged time steps for future covariates. Default is None.

        add_encoders : dict or None, optional
            Additional encoders for categorical variables. Default is None.

        quantile : float or None, optional
            The quantile level for prediction intervals. Default is 0.9 (90%).

        random_state : int or None, optional
            The random seed for reproducibility. Default is None.

        multi_models : bool, optional
            Whether to use multiple models. Default is True.

        use_static_covariates : bool, optional
            Whether to use static covariates. Default is True.

        categorical_past_covariates : str, list, or None, optional
            Categorical past covariates. Default is None.

        categorical_future_covariates : str, list, or None, optional
            Categorical future covariates. Default is None.

        categorical_static_covariates : str, list, or None, optional
            Categorical static covariates. Default is None.

        verbose : int, optional
            Verbosity level. Default is -1.

        linear_tree : bool, optional
            Whether to use linear tree models. Default is True.

        **darts_lightgbm_model_configs
            Additional configurations specific to the LightGBM model.
        """
        super().__init__(time_col=time_col, target_col=target_col)

        # Generate model configurations
        self.all_configs['model_configs'] = generate_function_kwargs(
            LGB,
            lags=lags,
            lags_past_covariates=lags_past_covariates,
            lags_future_covariates=lags_future_covariates,
            output_chunk_length=lags,
            add_encoders=add_encoders,
            random_state=random_state,
            multi_models=multi_models,
            use_static_covariates=use_static_covariates,
            categorical_past_covariates=categorical_past_covariates,
            categorical_future_covariates=categorical_future_covariates,
            categorical_static_covariates=categorical_static_covariates,
            verbose=verbose,
            linear_tree=linear_tree,
            **darts_lightgbm_model_configs
        )

        # Define and initialize the model
        self.model = self._define_model()

        # Update configurations
        self.all_configs.update(
            {
                'lags': lags,
                'quantile': quantile,
                'time_col': time_col,
                'target_col': target_col,
                'quantile_error': 0
            }
        )

    def _define_model(self):
        """Define the LightGBM model."""
        return LGB(**self.all_configs['model_configs'])


class XGBoostModel(DartsForecastMixin, GBDTModelMixin, IntervalEstimationMixin):
    @ParameterTypeAssert({
        'time_col': str,
        'target_col': str,
        'lags': int,
        'lags_past_covariates': (None, int),
        'lags_future_covariates': (None, int),
        'add_encoders': (dict, None),
        'quantile': (None, float),
        'random_state': (None, int),
        'multi_models': bool,
        'use_static_covariates': bool,
        'categorical_past_covariates': (str, list, None),
        'categorical_future_covariates': (str, list, None),
        'categorical_static_covariates': (str, list, None),
        'verbose': int
    }, 'XGBoostModel')
    def __init__(
            self,
            time_col,
            target_col,
            lags=1,
            lags_past_covariates=None,
            lags_future_covariates=None,
            add_encoders=None,
            quantile=0.9,
            random_state=None,
            multi_models=True,
            use_static_covariates=True,
            categorical_past_covariates=None,
            categorical_future_covariates=None,
            categorical_static_covariates=None,
            verbose=0,
            **darts_xgboost_model_configs
    ):
        """
        Initialize the XGBoostModel.

        Parameters
        ----------
        time_col : str
            The name of the time column in the input DataFrame.

        target_col : str
            The name of the target column in the input DataFrame.

        lags : int
            The number of lagged time steps to consider. Default is 1.

        lags_past_covariates : int or None, optional
            The number of lagged time steps for past covariates. Default is None.

        lags_future_covariates : int or None, optional
            The number of lagged time steps for future covariates. Default is None.

        add_encoders : dict or None, optional
            Additional encoders for categorical variables. Default is None.

        quantile : float or None, optional
            The quantile level for prediction intervals. Default is 0.9 (90%).

        random_state : int or None, optional
            The random seed for reproducibility. Default is None.

        multi_models : bool, optional
            Whether to use multiple models. Default is True.

        use_static_covariates : bool, optional
            Whether to use static covariates. Default is True.

        categorical_past_covariates : str, list, or None, optional
            Categorical past covariates. Default is None.

        categorical_future_covariates : str, list, or None, optional
            Categorical future covariates. Default is None.

        categorical_static_covariates : str, list, or None, optional
            Categorical static covariates. Default is None.

        verbose : int, optional
            Verbosity level. Default is 0.

        **darts_xgboost_model_configs
            Additional configurations specific to the XGBoost model.
        """
        super().__init__(time_col=time_col, target_col=target_col)

        # Generate model configurations
        self.all_configs['model_configs'] = generate_function_kwargs(
            XGB,
            lags=lags,
            lags_past_covariates=lags_past_covariates,
            lags_future_covariates=lags_future_covariates,
            output_chunk_length=lags,
            add_encoders=add_encoders,
            random_state=random_state,
            multi_models=multi_models,
            use_static_covariates=use_static_covariates,
            categorical_past_covariates=categorical_past_covariates,
            categorical_future_covariates=categorical_future_covariates,
            categorical_static_covariates=categorical_static_covariates,
            verbosity=verbose,
            **darts_xgboost_model_configs
        )

        # Define and initialize the model
        self.model = self._define_model()

        # Update configurations
        self.all_configs.update(
            {
                'lags': lags,
                'quantile': quantile,
                'time_col': time_col,
                'target_col': target_col,
                'quantile_error': 0
            }
        )

    def _define_model(self):
        """Define the XGBoost model."""
        return XGB(**self.all_configs['model_configs'])


class RandomForestModel(DartsForecastMixin, GBDTModelMixin, IntervalEstimationMixin):
    @ParameterTypeAssert({
        'time_col': str,
        'target_col': str,
        'lags': int,
        'lags_past_covariates': (None, int),
        'lags_future_covariates': (None, int),
        'add_encoders': (dict, None),
        'n_estimators': int,
        'quantile': (None, float),
        'random_state': (None, int),
        'multi_models': bool,
        'use_static_covariates': bool
    }, 'RandomForestModel')
    def __init__(
            self,
            time_col,
            target_col,
            lags=1,
            lags_past_covariates=None,
            lags_future_covariates=None,
            add_encoders=None,
            n_estimators=100,
            quantile=0.9,
            random_state=None,
            multi_models=True,
            use_static_covariates=True,
            **darts_random_forest_model_configs
    ):
        """
        Initialize the RandomForestModel.

        Parameters
        ----------
        time_col : str
            The name of the time column in the input DataFrame.

        target_col : str
            The name of the target column in the input DataFrame.

        lags : int
            The number of lagged time steps to consider.

        lags_past_covariates : int or None, optional
            The number of lagged time steps for past covariates. Default is None.

        lags_future_covariates : int or None, optional
            The number of lagged time steps for future covariates. Default is None.

        add_encoders : dict or None, optional
            Additional encoders for categorical variables. Default is None.

        n_estimators : int
            The number of trees in the forest.

        quantile : float or None, optional
            The quantile level for prediction intervals. Default is 0.9 (90%).

        random_state : int or None, optional
            The random seed for reproducibility. Default is None.

        multi_models : bool, optional
            Whether to use multiple models. Default is True.

        use_static_covariates : bool, optional
            Whether to use static covariates. Default is True.

        **darts_random_forest_model_configs
            Additional configurations specific to the RandomForest model.
        """
        super().__init__(time_col=time_col, target_col=target_col)

        # Generate model configurations
        self.all_configs['model_configs'] = generate_function_kwargs(
            RF,
            lags=lags,
            lags_past_covariates=lags_past_covariates,
            lags_future_covariates=lags_future_covariates,
            output_chunk_length=lags,
            add_encoders=add_encoders,
            n_estimators=n_estimators,
            random_state=random_state,
            multi_models=multi_models,
            use_static_covariates=use_static_covariates,
            **darts_random_forest_model_configs
        )

        # Define and initialize the model
        self.model = self._define_model()

        # Update configurations
        self.all_configs.update(
            {
                'lags': lags,
                'quantile': quantile,
                'time_col': time_col,
                'target_col': target_col,
                'quantile_error': 0
            }
        )

    def _define_model(self):
        """Define the RandomForest model."""
        return RF(**self.all_configs['model_configs'])

    @ParameterTypeAssert({
        'data': pd.DataFrame,
        'convert_dataframe_kwargs': (None, dict),
        'cv': int,
        'fit_kwargs': (None, dict)
    })
    def fit(self, data, cv=5, convert_dataframe_kwargs=None, fit_kwargs=None):
        """
        Fit the model to the provided data.

        Parameters
        ----------
        data : pd.DataFrame
            The input data.

        cv : int, optional
            The number of cross-validation folds. Default is 5.

        convert_dataframe_kwargs : dict or None, optional
            Additional keyword arguments for converting the DataFrame. Default is None.

        fit_kwargs : dict or None, optional
            Additional keyword arguments for fitting the model. Default is None.

        Returns
        -------
        self
            Returns an instance of the fitted model.
        """
        check_time_col_is_timestamp(data, self.all_configs['time_col'])

        super().fit(data=data, convert_dataframe_kwargs=convert_dataframe_kwargs,
                    fit_kwargs=fit_kwargs, valid_data=None)

        # Calculate quantile error if quantile is specified
        if self.all_configs['quantile'] is not None:
            self.all_configs['quantile_error'] = \
                self.calculate_confidence_interval_darts(data, fit_kwargs=fit_kwargs,
                                                         convert2dts_dataframe_kwargs=convert_dataframe_kwargs, cv=cv)

        return self



from PipelineTS.ml_model.gbdt import (
    CatBoostModel,
    LightGBMModel,
    XGBoostModel,
    RandomForestModel
)

from PipelineTS.ml_model.wide_gbrt import WideGBRTModel
from PipelineTS.ml_model.multi_output_model import (
    MultiOutputRegressorModel,
    MultiStepRegressorModel,
    RegressorChainModel
)



from tabulate import tabulate
from IPython.display import display

from spinesUtils.utils import (
    reindex_iterable_object,
    is_in_ipython
)
from spinesUtils import ParameterTypeAssert, ParameterValuesAssert

from PipelineTS.pipeline.pipeline_models import get_all_available_models
from PipelineTS.base.base_utils import drop_duplicates_with_order


MODELS = get_all_available_models()


def configs_check(configs):
    """
    Check the validity of the given configuration format.

    Parameters
    ----------
    configs : list
        List of configuration tuples for models.

    Returns
    -------
    valid : bool
        True if the configuration format is valid, False otherwise.

    Notes
    -----
    - Each tuple in the list must have 2 or 3 elements.
    - The last element of each tuple must be a dictionary.
    - The first element of each tuple must be a string or tuple of two strings.
    """
    condition_a = all(
        isinstance(i, tuple) and 2 <= len(i) <= 3 and isinstance(i[-1], dict) for i in configs
    )

    condition_b = all(
       isinstance(i[0], str) if len(i) == 2 else isinstance(i[0], str) and isinstance(i[1], str) for i in configs
    )

    return condition_a and condition_b


class PipelineConfigs:
    """
    Class to manage configuration details for models in the pipeline.

    Parameters
    ----------
    configs : list
        List of configuration tuples for models.

    Raises
    ------
    KeyError
        If model_configs does not contain required sub-configurations.
        If model_name is not found in the predefined MODELS.

    Notes
    -----
    - The configs parameter should be a list of tuples, where each tuple represents the configuration for a model.
    - Each tuple must have 2 or 3 elements. The last element must be a dictionary containing sub-configurations.
    - If the length of the tuple is 2, the first element must be a string.
    - If the length is 3, both the first and second elements must be strings.
    """
    @ParameterTypeAssert({
        'configs': list
    }, 'PipelineConfigs')
    @ParameterValuesAssert({
        'configs': configs_check
    }, 'PipelineConfigs')
    def __init__(self, configs):
        # Initialize sub-configurations
        self.sub_configs = {'init_configs': {}, 'fit_configs': {}, 'predict_configs': {}}

        # Remove duplicates and maintain order
        _to_process_configs = drop_duplicates_with_order(configs)

        self.configs = []
        for models_group in reindex_iterable_object(_to_process_configs, key=lambda s: s[0], index_start=1):
            for (index, (*model_name_list, model_configs)) in models_group:
                model_name = model_name_list[0]
                if len(model_name_list) == 2:
                    model_name_after_rename = model_name_list[1]
                else:
                    model_name_after_rename = f"{model_name}_{index}"

                # Check if required sub-configurations are present
                if not all(i in self.sub_configs for i in model_configs):
                    raise KeyError

                # Check if the model_name is predefined
                if model_name not in MODELS:
                    raise KeyError

                # Update with default sub-configurations if necessary
                if len(model_configs) < 3:
                    model_configs.update({k: self.sub_configs[k]
                                          for k in self.sub_configs if k not in model_configs})

                self.configs.append((model_name, model_name_after_rename, model_configs))

        # Display the configurations in a tabular format
        if is_in_ipython():
            display(
                tabulate(self.configs, headers=['model_name', 'model_name_after_rename', 'model_configs'],
                         tablefmt='html', colalign=("right", "left", "left"), showindex='always')
            )
        else:
            print(
                tabulate(self.configs, headers=['model_name', 'model_name_after_rename', 'model_configs'],
                         tablefmt='pretty', colalign=("right", "left", "left"), showindex='always')
            )

    def get_configs(self, model_name_after_rename):
        """
        Retrieve the configuration details for a specific model.

        Parameters
        ----------
        model_name_after_rename : str
            The name of the model for which to retrieve configuration details.

        Returns
        -------
        model_configs : dict or None
            A dictionary containing the configuration details for the specified model.

        Example
        -------
        >>> pipeline_configs = PipelineConfigs(configs=[('lightgbm', 'my_lightgbm', {'init_configs': {...}, 'fit_configs': {...}})])
        >>> configs_model_a = pipeline_configs.get_configs('my_lightgbm')
        >>> print(configs_model_a)
        {'init_configs': {...}, 'fit_configs': {...}, 'predict_configs': {}}
        """
        for (model_name, mnar, model_configs) in self.configs:
            if model_name_after_rename == mnar:
                return model_configs
        return None



from PipelineTS.nn_model.d_linear import DLinearModel
from PipelineTS.nn_model.n_linear import NLinearModel
from PipelineTS.nn_model.n_beats import NBeatsModel
from PipelineTS.nn_model.n_hits import NHitsModel
from PipelineTS.nn_model.tft import TFTModel
from PipelineTS.nn_model.gau import GAUModel
from PipelineTS.nn_model.stacking_rnn import StackingRNNModel
from PipelineTS.nn_model.time2vec import Time2VecModel
from PipelineTS.nn_model.transformer import TransformerModel
from PipelineTS.nn_model.tide import TiDEModel
from PipelineTS.nn_model.patch_rnn import PatchRNNModel
from PipelineTS.nn_model.tcn import TCNModel



import re

from lightgbm import LGBMRegressor
import numpy as np
import pandas as pd

from spinesTS.ml_model import MultiOutputRegressor as MOR, MultiStepRegressor as MSR
from sklearn.multioutput import RegressorChain
from spinesTS.preprocessing import split_series, lag_splits
from spinesUtils import generate_function_kwargs, ParameterTypeAssert, ParameterValuesAssert
from spinesUtils.asserts import raise_if_not
from spinesUtils.preprocessing import reshape_if
from spinesTS.pipeline import Pipeline

from PipelineTS.base.base import GBDTModelMixin, IntervalEstimationMixin
from PipelineTS.base.spines_base import SpinesMLModelMixin
from PipelineTS.utils import update_dict_without_conflict, check_time_col_is_timestamp


class _MultiOutputModelMixin(GBDTModelMixin, IntervalEstimationMixin, SpinesMLModelMixin):
    def __init__(
            self,
            time_col,
            target_col,
            lags=1,
            quantile=0.9,
            estimator=LGBMRegressor,
            differential_n=1,
            **model_configs
    ):
        """
        A mixin class for multi-output regression models using scikit-learn's ensemble.RandomForest.

        Parameters
        ----------
        time_col : str
            The column containing time information in the input data.
        target_col : str
            The column containing the target variable in the input data.
        lags : int, optional, default: 1
            The number of lagged values to use as input features for training and prediction.
        quantile : float, optional, default: 0.9
            The quantile used for interval prediction. Set to None for point prediction.
        estimator : sklearn.base.BaseEstimator, optional, default: LGBMRegressor
            The base estimator used for the multi-output regression model.
        differential_n : int,  optional, default: 1
            The number of differencing operations to apply to the target variable.
        **model_configs : dict
            Additional keyword arguments for configuring the base estimator.

        Attributes
        ----------
        _base_estimator : BaseEstimator
            The base estimator for the multi-output regression model.
        all_configs : dict
            A dictionary containing all configuration parameters for the model.
        last_dt : pandas.Timestamp or None
            The last timestamp in the input data.
        last_lags_dataframe : pandas.DataFrame or None
            The DataFrame containing the last lagged values used for prediction.
        x : pandas.DataFrame or None
            The last input features used for training.
        """
        super().__init__(time_col=time_col, target_col=target_col)

        self._base_estimator = estimator
        self.all_configs['model_configs'] = generate_function_kwargs(
            self._base_estimator,
            **model_configs
        )

        if 'LGBMRegressor' in re.split("<|>|class| |\.|\'", str(estimator)):
            self.all_configs['model_configs'] = update_dict_without_conflict(self.all_configs['model_configs'], {
                'verbose': -1
            })

        self.last_dt = None
        self.last_lags_dataframe = None

        self.all_configs.update(
            {
                'lags': lags,
                'quantile': quantile,
                'time_col': time_col,
                'target_col': target_col,
                'quantile_error': 0,
                'differential_n': differential_n
            }
        )

        self.x = None

    def _define_model(self):
        """
        Define the multi-output regression model using scikit-learn's ensemble.RandomForest.

        Raises
        ------
        NotImplementedError
            This method should be implemented by subclasses.
        """
        raise NotImplementedError

    @ParameterTypeAssert({
        'mode': str
    })
    @ParameterValuesAssert({
        'mode': ('train', 'predict')
    })
    def _data_preprocess(self, data, mode='train', update_last_dt=False):
        """
        Preprocess the input data for training or prediction.

        Parameters
        ----------
        data : pandas.DataFrame
            The input data.
        mode : {'train', 'predict'}, optional, default: 'train'
            The mode indicating whether to preprocess data for training or prediction.
        update_last_dt : bool, optional, default: False
            Whether to update the last timestamp attribute.

        Returns
        -------
        tuple or numpy.ndarray
            If 'mode' is 'train', returns a tuple (x_train, y_train).
            If 'mode' is 'predict', returns lagged splits of the target column.
        """
        data[self.all_configs['time_col']] = pd.to_datetime(data[self.all_configs['time_col']])
        if update_last_dt:
            self.last_dt = data[self.all_configs['time_col']].max()

        if mode == 'train':
            # x_train, y_train
            return split_series(data[self.all_configs['target_col']], data[self.all_configs['target_col']],
                                window_size=self.all_configs['lags'], pred_steps=self.all_configs['lags'])
        else:
            return lag_splits(data[self.all_configs['target_col']], window_size=self.all_configs['lags'])

    def fit(self, data, cv=5, fit_kwargs=None):
        """
        Fit the multi-output regression model to the training data.

        Parameters
        ----------
        data : pandas.DataFrame
            The training data.
        cv : int, optional, default: 5
            The number of cross-validation folds.
        fit_kwargs : dict or None, optional, default: None
            Additional keyword arguments for the fitting process.

        Returns
        -------
        self
            Returns an instance of the fitted model.
        """
        check_time_col_is_timestamp(data, self.all_configs['time_col'])

        data = data[[self.all_configs['time_col'], self.all_configs['target_col']]]

        self.last_lags_dataframe = data.iloc[-(2 * self.all_configs['lags'] + 1):, :]

        if fit_kwargs is None:
            fit_kwargs = {}

        x, y = self._data_preprocess(data, update_last_dt=True, mode='train')
        x = pd.DataFrame(x)

        self.x = pd.DataFrame(self._data_preprocess(data, mode='predict')).iloc[-1:, :]

        # difference
        x_after_diff = np.diff(x, n=self.all_configs['differential_n'], axis=1)
        self.model.fit(x_after_diff, y, **fit_kwargs)

        if self.all_configs['quantile'] is not None:
            self.all_configs['quantile_error'] = \
                self.calculate_confidence_interval_mor(data, fit_kwargs=fit_kwargs, cv=cv)

        return self

    def _extend_predict(self, x, n, predict_kwargs):
        """
        Extrapolation prediction.

        Parameters
        ----------
        x : numpy.ndarray
            Data to predict, must be 2-dimensional.
        n : int
            Number of prediction steps.
        predict_kwargs : dict
            Additional keyword arguments for the prediction process.

        Returns
        -------
        list
            List of predictions.
        """
        assert isinstance(n, int)
        assert x.ndim == 2

        x_after_diff = np.diff(x, n=self.all_configs['differential_n'], axis=1)
        current_res = self.model.predict(x_after_diff, **predict_kwargs)
        current_res = reshape_if(current_res, current_res.ndim == 1, (1, -1))

        if n is None:
            return current_res.squeeze().tolist()
        elif n <= current_res.shape[1]:
            return current_res.squeeze().tolist()[:n]
        else:
            res = current_res.squeeze().tolist()

            for i in range(n - self.all_configs['lags']):
                x = np.concatenate((x[:, 1:], current_res[:, 0:1]), axis=1)
                
                x_after_diff = np.diff(x, n=self.all_configs['differential_n'], axis=1)
                current_res = self.model.predict(x_after_diff, **predict_kwargs)
                current_res = reshape_if(current_res, current_res.ndim == 1, (1, -1))

                res.append(current_res.squeeze().tolist()[-1])

            return res

    def predict(self, n, data=None, predict_kwargs=None):
        """
        Predict future values using the trained model.

        Parameters
        ----------
        n : int
            Number of prediction steps.
        data : pandas.DataFrame or None, optional, default: None
            Additional data for prediction.
        predict_kwargs : dict or None, optional, default: None
            Additional keyword arguments for the prediction process.

        Returns
        -------
        pandas.DataFrame
            DataFrame containing the predicted values and corresponding timestamps.
        """
        if predict_kwargs is None:
            predict_kwargs = {}

        if data is not None:
            check_time_col_is_timestamp(data, self.all_configs['time_col'])

            raise_if_not(
                ValueError, len(data) >= self.all_configs['lags'],
                'The length of the series must be greater than or equal to the lags.'
            )

            x = self._data_preprocess(
                data, mode='predict', update_last_dt=False
            )[-1, :]
            x = reshape_if(x, x.ndim == 1, (1, -1))
            last_dt = data[self.all_configs['time_col']].max()
        else:
            x = reshape_if(self.x.values, self.x.values.ndim == 1, (1, -1))
            last_dt = self.last_dt

        res = self._extend_predict(x, n, predict_kwargs=predict_kwargs)  # list

        assert len(res) == n

        res = pd.DataFrame(res, columns=[self.all_configs['target_col']])
        res[self.all_configs['time_col']] = \
            last_dt + pd.to_timedelta(range(res.index.shape[0] + 1), unit='D')[1:]

        if self.all_configs['quantile'] is not None:
            res = self.interval_predict(res)

        return self.chosen_cols(res)


class MultiOutputRegressorModel(_MultiOutputModelMixin):
    def __init__(
            self,
            time_col,
            target_col,
            lags=1,
            quantile=0.9,
            estimator=LGBMRegressor,
            **model_configs
    ):
        """
        Multi-output regression model using scikit-learn's ensemble.RandomForest with a regressor chain.

        Parameters
        ----------
        time_col : str
            The column containing time information in the input data.
        target_col : str
            The column containing the target variable in the input data.
        lags : int, optional, default: 1
            The number of lagged values to use as input features for training and prediction.
        quantile : float, optional, default: 0.9
            The quantile used for interval prediction. Set to None for point prediction.
        estimator : sklearn.base.BaseEstimator, optional, default: LGBMRegressor
            The base estimator used for the multi-output regression model.
        **model_configs : dict
            Additional keyword arguments for configuring the base estimator.

        Attributes
        ----------
        model : spinesTS.pipeline.Pipeline
            The pipeline containing the multi-output regressor model.
        """
        super().__init__(
            time_col=time_col,
            target_col=target_col,
            lags=lags,
            quantile=quantile,
            estimator=estimator,
            **model_configs)

        self.model = self._define_model()

    def _define_model(self):
        """
        Define the multi-output regressor model using a regressor chain.

        Returns
        -------
        spinesTS.pipeline.Pipeline
            The pipeline containing the multi-output regressor model.
        """
        return Pipeline([
            ('model', MOR(self._base_estimator(**self.all_configs['model_configs'])))
        ])


class MultiStepRegressorModel(_MultiOutputModelMixin):
    def __init__(
            self,
            time_col,
            target_col,
            lags=1,
            quantile=0.9,
            estimator=LGBMRegressor,
            **model_configs
    ):
        """
        Multi-step regression model using scikit-learn's ensemble.RandomForest with a multi-step regressor.

        Parameters
        ----------
        time_col : str
            The column containing time information in the input data.
        target_col : str
            The column containing the target variable in the input data.
        lags : int, optional, default: 1
            The number of lagged values to use as input features for training and prediction.
        quantile : float, optional, default: 0.9
            The quantile used for interval prediction. Set to None for point prediction.
        estimator : sklearn.base.BaseEstimator, optional, default: LGBMRegressor
            The base estimator used for the multi-step regression model.
        **model_configs : dict
            Additional keyword arguments for configuring the base estimator.

        Attributes
        ----------
        model : spinesTS.pipeline.Pipeline
            The pipeline containing the multi-step regressor model.
        """
        super().__init__(
            time_col=time_col,
            target_col=target_col,
            lags=lags,
            quantile=quantile,
            estimator=estimator,
            **model_configs)

        self.model = self._define_model()

    def _define_model(self):
        """
        Define the multi-step regressor model using a multi-step regressor.

        Returns
        -------
        spinesTS.pipeline.Pipeline
            The pipeline containing the multi-step regressor model.
        """
        return Pipeline([
            ('model', MSR(self._base_estimator(**self.all_configs['model_configs'])))
        ])


class RegressorChainModel(_MultiOutputModelMixin):
    def __init__(
            self,
            time_col,
            target_col,
            lags=1,
            quantile=0.9,
            estimator=LGBMRegressor,
            **model_configs
    ):
        """
        Regressor chain model using scikit-learn's ensemble.RandomForest.

        Parameters
        ----------
        time_col : str
            The column containing time information in the input data.
        target_col : str
            The column containing the target variable in the input data.
        lags : int, optional, default: 1
            The number of lagged values to use as input features for training and prediction.
        quantile : float, optional, default: 0.9
            The quantile used for interval prediction. Set to None for point prediction.
        estimator : sklearn.base.BaseEstimator, optional, default: LGBMRegressor
            The base estimator used for the regressor chain model.
        **model_configs : dict
            Additional keyword arguments for configuring the base estimator.

        Attributes
        ----------
        model : sklearn.multioutput.RegressorChain
            The regressor chain model.
        """
        super().__init__(
            time_col=time_col,
            target_col=target_col,
            lags=lags,
            quantile=quantile,
            estimator=estimator,
            **model_configs)

        self.model = self._define_model()

    def _define_model(self):
        """
        Define the regressor chain model using a base estimator.

        Returns
        -------
        sklearn.multioutput.RegressorChain
            The regressor chain model.
        """
        return RegressorChain(self._base_estimator(**self.all_configs['model_configs']))



from frozendict import frozendict

from PipelineTS.statistic_model import *
from PipelineTS.ml_model import *
from PipelineTS.nn_model import *

from PipelineTS.base.base_utils import get_model_name_before_initial


def get_all_available_models():
    """
    Retrieve a dictionary of all available model classes in the pipeline.

    Returns
    -------
    models : frozendict
        A frozendict containing model names as keys and corresponding model class references as values.

    Notes
    -----
    - The function attempts to import external dependencies to check for additional models.
    - If the 'prophet' package is installed, a 'prophet' model will be added to the available models.
    """
    try:
        from prophet import Prophet
        extra_pkg_installed = True
    except ImportError:
        extra_pkg_installed = False

    MODELS = frozendict({
        'auto_arima': AutoARIMAModel,
        'catboost': CatBoostModel,
        'lightgbm': LightGBMModel,
        'xgboost': XGBoostModel,
        'wide_gbrt': WideGBRTModel,
        'd_linear': DLinearModel,
        'n_linear': NLinearModel,
        'n_beats': NBeatsModel,
        'n_hits': NHitsModel,
        'tcn': TCNModel,
        'tft': TFTModel,
        'gau': GAUModel,
        'stacking_rnn': StackingRNNModel,
        'time2vec': Time2VecModel,
        'multi_output_model': MultiOutputRegressorModel,
        'multi_step_model': MultiStepRegressorModel,
        'transformer': TransformerModel,
        'random_forest': RandomForestModel,
        'tide': TiDEModel,
        'patch_rnn': PatchRNNModel,
        'regressor_chain': RegressorChainModel
    })

    if extra_pkg_installed:
        MODELS = MODELS.set('prophet', ProphetModel)

    return MODELS


def get_all_model_class_name():
    models = dict(get_all_available_models())

    res = {}

    for k, v in models.items():
        res[get_model_name_before_initial(v)] = v

    return frozendict(res)


# utility functions
import numpy as np


def load_dataset_to_darts(
        data,
        time_col,
        target_col,
        fill_missing_dates=False,
        freq=None,
        fillna_value=None,
        static_covariates=None,
        hierarchy=None
):
    from darts.timeseries import TimeSeries

    return TimeSeries.from_dataframe(
        data,
        time_col=time_col,
        value_cols=target_col,
        fill_missing_dates=fill_missing_dates,
        freq=freq,
        fillna_value=fillna_value,
        static_covariates=static_covariates,
        hierarchy=hierarchy
    )


def update_dict_without_conflict(dict_a, dict_b):
    for i in dict_b:
        if i not in dict_a:
            dict_a[i] = dict_b[i]
    return dict_a


def check_time_col_is_timestamp(data, time_col):
    from spinesUtils.asserts import raise_if_not

    raise_if_not(TypeError, data[time_col].dtype == 'datetime64[ns]',
                 'The time column must be of type datetime64[ns], '
                 'consider use pandas.to_datetime to convert it.')


def compute_time_interval(data, time_col):
    return data[time_col].diff().mode().values[0]


def time_diff(a, b):
    return np.timedelta64(a - b)



import logging
import torch
from darts.models.forecasting.dlinear import DLinearModel as d_linear
from spinesUtils.asserts import generate_function_kwargs
from PipelineTS.base.base import NNModelMixin, IntervalEstimationMixin
from PipelineTS.base.darts_base import DartsForecastMixin
from PipelineTS.utils import update_dict_without_conflict

# Suppressing unnecessary PyTorch Lightning warnings
logging.getLogger("pytorch_lightning.utilities.rank_zero").setLevel(logging.WARNING)
logging.getLogger("pytorch_lightning.accelerators.cuda").setLevel(logging.WARNING)


class DLinearModel(DartsForecastMixin, NNModelMixin, IntervalEstimationMixin):
    def __init__(
            self,
            time_col,
            target_col,
            lags=6,
            shared_weights=False,
            kernel_size=25,
            const_init=True,
            use_static_covariates=True,
            loss_fn=torch.nn.MSELoss(),
            torch_metrics=None,
            optimizer_cls=torch.optim.Adam,
            optimizer_kwargs=None,
            lr_scheduler_cls=None,
            lr_scheduler_kwargs=None,
            use_reversible_instance_norm=False,
            batch_size=32,
            n_epochs=100,
            nr_epochs_val_period=1,
            add_encoders=None,
            enable_progress_bar=False,
            enable_model_summary=False,
            pl_trainer_kwargs=None,
            quantile=0.9,
            random_state=None,
            accelerator=None
    ):
        """
        DLinearModel: A wrapper for Darts' DLinear forecasting model with additional features.

        Parameters
        ----------
        time_col : str
            The column containing time information in the input data.
        target_col : str
            The column containing the target variable in the input data.
        lags : int, optional, default: 6
            The number of lagged values to use as input features for training and prediction.
        shared_weights : bool, optional, default: False
            Whether to share weights across all inputs in the DLinear model.
        kernel_size : int, optional, default: 25
            The size of the kernel in the DLinear model.
        const_init : bool, optional, default: True
            Whether to use constant initialization for weights in the DLinear model.
        use_static_covariates : bool, optional, default: True
            Whether to use static covariates in the DLinear model.
        loss_fn : torch.nn.Module, optional, default: torch.nn.MSELoss
            The loss function used during training.
        torch_metrics : torch.nn.Module or None, optional, default: None
            Additional metrics for training evaluation.
        optimizer_cls : torch.optim.Optimizer, optional, default: torch.optim.Adam
            The optimizer used for training.
        optimizer_kwargs : dict or None, optional, default: None
            Additional keyword arguments for configuring the optimizer.
        lr_scheduler_cls : torch.optim.lr_scheduler, optional, default: None
            The learning rate scheduler used during training.
        lr_scheduler_kwargs : dict or None, optional, default: None
            Additional keyword arguments for configuring the learning rate scheduler.
        use_reversible_instance_norm : bool, optional, default: False
            Whether to use reversible instance normalization in the DLinear model.
        batch_size : int, optional, default: 32
            The batch size used during training.
        n_epochs : int, optional, default: 100
            The number of epochs for training.
        nr_epochs_val_period : int, optional, default: 1
            The number of epochs between each validation period during training.
        add_encoders : list or None, optional, default: None
            Additional encoders for the DLinear model.
        enable_progress_bar : bool, optional, default: False
            Whether to enable the progress bar during training.
        enable_model_summary : bool, optional, default: False
            Whether to enable the model summary during training.
        pl_trainer_kwargs : dict or None, optional, default: None
            Additional keyword arguments for configuring the PyTorch Lightning trainer.
        quantile : float, optional, default: 0.9
            The quantile used for interval prediction. Set to None for point prediction.
        random_state : int or None, optional, default: None
            The random seed for reproducibility.
        accelerator : str or None, optional, default: None
            The PyTorch Lightning accelerator to use during training.

        Attributes
        ----------
        model : darts.models.forecasting.dlinear.DLinearModel
            The Darts DLinear forecasting model.
        """
        super().__init__(time_col=time_col, target_col=target_col, accelerator=accelerator)

        if pl_trainer_kwargs is None:
            pl_trainer_kwargs = {}

        pl_trainer_kwargs = update_dict_without_conflict(pl_trainer_kwargs, {
            'accelerator': self.accelerator,
            'enable_progress_bar': enable_progress_bar,
            'enable_model_summary': enable_model_summary
        })

        self.all_configs['model_configs'] = generate_function_kwargs(
            d_linear,
            input_chunk_length=lags,
            output_chunk_length=lags,
            shared_weights=shared_weights,
            kernel_size=kernel_size,
            const_init=const_init,
            use_static_covariates=use_static_covariates,
            loss_fn=loss_fn,
            torch_metrics=torch_metrics,
            optimizer_cls=optimizer_cls,
            optimizer_kwargs=optimizer_kwargs,
            lr_scheduler_cls=lr_scheduler_cls,
            lr_scheduler_kwargs=lr_scheduler_kwargs,
            use_reversible_instance_norm=use_reversible_instance_norm,
            batch_size=batch_size,
            n_epochs=n_epochs,
            nr_epochs_val_period=nr_epochs_val_period,
            add_encoders=add_encoders,
            pl_trainer_kwargs=pl_trainer_kwargs,
            random_state=random_state,
        )
        self.model = self._define_model()

        self.all_configs.update(
            {
                'lags': lags,
                'quantile': quantile,
                'time_col': time_col,
                'target_col': target_col,
                'enable_progress_bar': enable_progress_bar,
                'enable_model_summary': enable_model_summary
            }
        )

    def _define_model(self):
        """
        Define the DLinear forecasting model.

        Returns
        -------
        darts.models.forecasting.dlinear.DLinearModel
            The Darts DLinear forecasting model.
        """
        return d_linear(**self.all_configs['model_configs'])



from darts.models import AutoARIMA
from spinesUtils import generate_function_kwargs

from PipelineTS.base.base import StatisticModelMixin, IntervalEstimationMixin
from PipelineTS.base.darts_base import DartsForecastMixin
from PipelineTS.utils import check_time_col_is_timestamp


class AutoARIMAModel(DartsForecastMixin, StatisticModelMixin, IntervalEstimationMixin):
    def __init__(
            self,
            time_col,
            target_col,
            lags=1,
            start_p=8,
            max_p=12,
            start_q=1,
            seasonal=False,
            quantile=0.9,
            seasonal_length=12,
            n_jobs=-1,
            **darts_auto_arima_configs
    ):
        """
        AutoARIMAModel: A wrapper for the AutoARIMA model from the darts library with additional features.

        Parameters
        ----------
        time_col : str
            The column containing time information in the input data.
        target_col : str
            The column containing the target variable in the input data.
        lags : int, optional, default: 1
            The number of lagged values to use as input features for training and prediction.
        start_p : int, optional, default: 8
            The starting value for the order of autoregressive (AR) component.
        max_p : int, optional, default: 12
            The maximum value for the order of autoregressive (AR) component.
        start_q : int, optional, default: 1
            The starting value for the order of moving average (MA) component.
        seasonal : bool, optional, default: False
            Whether the time series exhibits seasonality.
        seasonal_length : int, optional, default: 12
            The length of the seasonal cycle, if seasonality is True.
        quantile : float, optional, default: 0.9
            The quantile used for interval prediction. Set to None for point prediction.
        n_jobs : int, optional, default: -1
            The number of jobs to run in parallel during model fitting.
        **darts_auto_arima_configs
            Additional keyword arguments for configuring the AutoARIMA model.

        Attributes
        ----------
        model : darts.models.AutoARIMA
            The AutoARIMA model from the darts library.
        """
        super().__init__(time_col=time_col, target_col=target_col)

        self.all_configs['model_configs'] = generate_function_kwargs(
            AutoARIMA,
            start_p=start_p,
            max_p=max_p,
            start_q=start_q,
            seasonal=seasonal,
            seasonal_length=seasonal_length,
            n_jobs=n_jobs,
            **darts_auto_arima_configs
        )

        self.model = self._define_model()

        self.all_configs.update(
            {
                'lags': lags,   # meanness, but only to follow coding conventions
                'quantile': quantile,
                'time_col': time_col,
                'target_col': target_col,
                'quantile_error': 0
            }
        )

    def _define_model(self):
        """
        Define the AutoARIMA model from the darts library.

        Returns
        -------
        darts.models.AutoARIMA
            The AutoARIMA model from the darts library.
        """
        return AutoARIMA(**self.all_configs['model_configs'])

    def fit(self, data, convert_dataframe_kwargs=None, cv=5, fit_kwargs=None):
        """
        Fit the AutoARIMA model on the input data.

        Parameters
        ----------
        data : pd.DataFrame
            The input data in pandas DataFrame format.
        convert_dataframe_kwargs : None, optional, default: None
            Additional keyword arguments for converting the input data to the required format.
        cv : int, optional, default: 5
            The number of cross-validation folds.
        fit_kwargs : None, optional, default: None
            Additional keyword arguments for fitting the model.

        Returns
        -------
        self : AutoARIMAModel
            Returns the instance itself.
        """
        check_time_col_is_timestamp(data, self.all_configs['time_col'])

        super().fit(
            data,
            convert_dataframe_kwargs=convert_dataframe_kwargs,
            fit_kwargs=fit_kwargs
        )

        if self.all_configs['quantile'] is not None:
            self.all_configs['quantile_error'] = \
                self.calculate_confidence_interval_darts(data, fit_kwargs=fit_kwargs,
                                                         convert2dts_dataframe_kwargs=convert_dataframe_kwargs, cv=cv)

        return self

    def predict(self, n, **kwargs):
        """
        Make predictions using the fitted AutoARIMA model.

        Parameters
        ----------
        n : int
            The number of time steps to predict.
        kwargs : dict
            Additional keyword arguments for making predictions.
        """
        res = super().predict(n, **kwargs)
        res = self.rename_prediction(res)

        if self.all_configs['quantile'] is not None:
            res = self.interval_predict(res)

        return self.chosen_cols(res)



import numpy as np

from spinesUtils.asserts import ParameterTypeAssert


@ParameterTypeAssert({
    'yt': np.ndarray,
    'left_pred': np.ndarray,
    'right_pred': np.ndarray
})
def quantile_acc(yt: np.ndarray, left_pred: np.ndarray, right_pred: np.ndarray) -> float:
    """
    Calculate the accuracy of prediction intervals.

    Parameters
    ----------
    yt : np.ndarray
        The true values.

    left_pred : np.ndarray
        The left bound of the prediction interval.

    right_pred : np.ndarray
        The right bound of the prediction interval.

    Returns
    -------
    float
        The accuracy of the prediction intervals, computed as the ratio of correct predictions to the total number of samples.
    """
    return np.sum((yt >= left_pred) * (yt <= right_pred)) / yt.shape[0]



from sklearn.preprocessing import MinMaxScaler, QuantileTransformer, StandardScaler
from spinesTS.preprocessing import GaussRankScaler


class Scaler:
    def __init__(self, scaler_name='min_max'):
        """
        Initialize a scaler object.

        Parameters
        ----------
        scaler_name : str, default='min_max'
            The name of the scaler. It can be 'min_max', 'gauss_rank', 'quantile', or 'standard'.
            - 'min_max': MinMaxScaler from sklearn.preprocessing.
            - 'gauss_rank': GaussRankScaler from spinesTS.preprocessing.
            - 'quantile': QuantileTransformer from sklearn.preprocessing.
            - 'standard': StandardScaler from sklearn.preprocessing.
        """
        self.scaler_name = scaler_name

        if self.scaler_name == 'min_max':
            self.scaler = MinMaxScaler()
        elif self.scaler_name == 'gauss_rank':
            self.scaler = GaussRankScaler()
        elif self.scaler_name == 'quantile':
            self.scaler = QuantileTransformer()
        elif self.scaler_name == 'standard':
            self.scaler = StandardScaler()
        else:
            raise ValueError('Unknown scaler name')

    def fit(self, X):
        """
        Fit the scaler to the input data.

        Parameters
        ----------
        X : array-like or pd.DataFrame
            Input data.

        Returns
        -------
        self : Scaler
            The fitted scaler object.
        """
        self.scaler.fit(X)
        return self

    def fit_transform(self, X):
        """
        Fit the scaler to the input data and transform it.

        Parameters
        ----------
        X : array-like or pd.DataFrame
            Input data.

        Returns
        -------
        transformed_data : array-like
            The transformed data.
        """
        return self.scaler.fit_transform(X)

    def transform(self, X):
        """
        Transform the input data using the fitted scaler.

        Parameters
        ----------
        X : array-like or pd.DataFrame
            Input data.

        Returns
        -------
        transformed_data : array-like
            The transformed data.
        """
        return self.scaler.transform(X)

    def inverse_transform(self, X):
        """
        Inverse transform the input data using the fitted scaler.

        Parameters
        ----------
        X : array-like or pd.DataFrame
            Transformed data.

        Returns
        -------
        original_data : array-like
            The original data before scaling.
        """
        return self.scaler.inverse_transform(X)



import datetime
from copy import deepcopy


def _get_object_name(model_obj):
    """
    Extracts and returns the object name from its string representation.

    Parameters
    ----------
    model_obj : object
        The object for which the name needs to be extracted.

    Returns
    -------
    str
        The extracted object name.
    """
    import re
    return re.split('<|>', str(model_obj))[1].split(' ')[0].strip().split('.')[-1]


def _hash_string(string):
    """
    Hashes a given string using MD5 algorithm.

    Parameters
    ----------
    string : str
        The input string to be hashed.

    Returns
    -------
    str
        The first 12 characters of the hashed string.
    """
    import hashlib
    import random
    import time

    # Create an MD5 hash object
    hash_object = hashlib.md5()

    # Update the hash object's value with the string to be encrypted
    hash_object.update((string + str(random.random()) + str(time.time())).encode('utf-8'))

    # Get the encrypted result
    encrypted_string = hash_object.hexdigest()

    return encrypted_string[:12]


def _zip_file(zipfile_fp, *file_fp):
    """
    Compresses multiple files into a zip archive.

    Parameters
    ----------
    zipfile_fp : str
        The path to the zip file.

    *file_fp : str
        Variable length list of file paths to be compressed.
    """
    import zipfile
    from pathlib import Path

    from spinesUtils.asserts import raise_if_not

    raise_if_not(ValueError, zipfile_fp.endswith('.zip'), "`zipfile_fp` must be a string with the `.zip` suffix")

    with zipfile.ZipFile(zipfile_fp, 'w') as zipf:
        for file in file_fp:
            zipf.write(file, Path(file).name)


def _load_zip_file(zipfile_fp):
    """
    Extracts the contents of a zip file to a temporary directory.

    Parameters
    ----------
    zipfile_fp : str
        The path to the zip file.

    Returns
    -------
    str
        The path to the temporary directory containing the extracted contents.
    """
    import zipfile
    from pathlib import Path

    from spinesUtils.asserts import raise_if, raise_if_not
    raise_if(ValueError, Path(zipfile_fp).is_dir(), "`zipfile_fp` must be a file name, not a directory.")

    raise_if_not(ValueError, zipfile_fp.strip().endswith('.zip'),
                 "`zipfile_fp` must be a string with the `.zip` suffix")

    # To ensure that folder names are as unique as possible
    tmp_unzip_fp = str(Path(zipfile_fp).parent.absolute().joinpath(
        f'PIPELINETS_MODEL_{_hash_string(zipfile_fp)}_{int(datetime.datetime.now().timestamp()*1e6)}/'))

    with zipfile.ZipFile(zipfile_fp, 'r') as zip_ref:
        zip_ref.extractall(tmp_unzip_fp)

    return tmp_unzip_fp


def _save_single_model(path, model, scaler=None):
    """
    Save a machine learning model along with its scaler to a zip file.

    Parameters
    ----------
    path : str
        The path to the zip file.

    model : object
        The machine learning model to be saved.

    scaler : object, optional
        The scaler associated with the model.

    Returns
    -------
    str
        The path to the saved zip file.
    """
    from pathlib import Path
    import shutil

    import cloudpickle
    from spinesUtils.asserts import raise_if, raise_if_not

    from PipelineTS.base import DartsForecastMixin, NNModelMixin

    raise_if(ValueError, Path(path).is_dir(), "`path` must be a file name, not a directory.")
    raise_if_not(ValueError, path.endswith('.zip'), "`path` must be a string with the `.zip` suffix")

    path = str(Path(path).absolute())

    zipfile_fp, pkl_file_fp, darts_model_weights_fp = path, path.strip()[:-4] + '.pkl', None

    model_fp = Path(path.strip()[:-4] + '/')

    # Create model directory
    Path.mkdir(model_fp)

    pkl_file_fp = str(model_fp.joinpath(Path(path.strip()[:-4] + '.pkl').name))

    if isinstance(model, DartsForecastMixin) and isinstance(model, NNModelMixin):
        model.model.save(str(model_fp.joinpath(_get_object_name(model) + '.pt')))

    with open(pkl_file_fp, 'wb') as f:
        if scaler is not None:
            cloudpickle.dump([model, scaler], f)
        else:
            cloudpickle.dump(model, f)

    if isinstance(model, DartsForecastMixin) and isinstance(model, NNModelMixin):
        _zip_file(zipfile_fp, pkl_file_fp, model_fp.joinpath(_get_object_name(model) + '.pt'),
                  model_fp.joinpath(_get_object_name(model) + '.pt.ckpt'))
    else:
        _zip_file(zipfile_fp, pkl_file_fp)

    shutil.rmtree(model_fp)

    return zipfile_fp


def _load_single_model(path, unzip_file_path=None, unzip=True):
    """
    Load a machine learning model from a zip file.

    Parameters
    ----------
    path : str
        The path to the zip file.

    unzip_file_path : str, optional
        The path to a pre-extracted directory containing the contents of the zip file.

    unzip : bool, optional
        If True, the zip file is extracted to a temporary directory.

    Returns
    -------
    tuple or object
        If a scaler is saved along with the model, returns a tuple (model, scaler). Otherwise, returns the model.
    """
    import os
    from pathlib import Path
    import shutil

    import cloudpickle
    from spinesUtils.asserts import raise_if, raise_if_not

    from PipelineTS.base import DartsForecastMixin, NNModelMixin

    raise_if(ValueError, Path(path).is_dir(), "`path` must be a file name, not a directory.")
    raise_if_not(ValueError, path.endswith('.zip'), "`path` must be a string with the `.zip` suffix")
    raise_if(ValueError, unzip_file_path is None and unzip is False,
             "`unzip_file_path` must be specified when `unzip` is False.")
    raise_if(ValueError, unzip_file_path is not None and unzip is True,
             "`unzip_file_path` must be None when `unzip` is True.")

    if unzip:
        unzip_file_fp = _load_zip_file(path)
    else:
        unzip_file_fp = unzip_file_path

    raise_if_not(ValueError, any([i.endswith('.pkl') for i in os.listdir(unzip_file_fp)]),
                 "Zip file must contain one file with the `.pkl` suffix.")

    model = None
    scaler = None

    for i in os.listdir(unzip_file_fp):
        if i.endswith('.pkl'):
            with open(str(Path(unzip_file_fp).joinpath(i)), 'rb') as f:
                model = cloudpickle.load(f)

            if isinstance(model, list) and len(model) == 2:
                (model, scaler) = model

    if isinstance(model, DartsForecastMixin) and isinstance(model, NNModelMixin):
        model.model = model.model.load(
            str(Path(unzip_file_fp).joinpath(_get_object_name(model) + '.pt'))
        )

    shutil.rmtree(unzip_file_fp)

    if scaler is not None:
        return model, scaler

    return model


def _save_pipeline(path, model):
    """
    Save a machine learning pipeline to a zip file.

    Parameters
    ----------
    path : str
        The path to the zip file.

    model : object
        The machine learning pipeline to be saved.
    """
    from pathlib import Path
    import shutil

    import cloudpickle
    from spinesUtils.asserts import raise_if, raise_if_not

    raise_if(ValueError, Path(path).is_dir(), "`path` must be a file name, not a directory.")
    raise_if_not(ValueError, path.endswith('.zip'), "`path` must be a string with the `.zip` suffix")

    zipfile_fp, darts_model_weights_fp = path, None

    # Make a directory with the pkl file name
    name_subfix = _hash_string(path.strip()[:-4])
    pipeline_path = Path(path).parent.joinpath(path.strip()[:-4] + name_subfix + '/')
    Path.mkdir(pipeline_path)

    pkl_file_fp = str(Path(pipeline_path).joinpath('pipeline.pkl'))

    file_fps = [pkl_file_fp]
    for (sub_model_name, sub_model) in model.models_:
        mpath = str(pipeline_path.joinpath(sub_model_name + '.zip'))
        file_fps.append(mpath)
        _save_single_model(mpath, sub_model)

    with open(pkl_file_fp, 'wb') as f:
        new_model = deepcopy(model)
        new_model.models_ = []
        new_model.best_model_ = None
        cloudpickle.dump(new_model, f)

    _zip_file(zipfile_fp, *file_fps)

    shutil.rmtree(pipeline_path)

    return zipfile_fp


def _load_pipeline(path, unzip_file_path=None, unzip=True):
    """
    Load a machine learning pipeline from a zip file.

    Parameters
    ----------
    path : str
        The path to the zip file.

    unzip_file_path : str, optional
        The path to a pre-extracted directory containing the contents of the zip file.

    unzip : bool, optional
        If True, the zip file is extracted to a temporary directory.

    Returns
    -------
    object
        The loaded machine learning pipeline.
    """
    import os
    from pathlib import Path
    import shutil

    import cloudpickle
    from spinesUtils.asserts import raise_if, raise_if_not

    raise_if(ValueError, Path(path).is_dir(), "`path` must be a file name, not a directory.")
    raise_if_not(ValueError, path.endswith('.zip'), "`path` must be a string with the `.zip` suffix")
    raise_if(ValueError, unzip_file_path is None and unzip is False,
             "`unzip_file_path` must be specified when `unzip` is False.")
    raise_if(ValueError, unzip_file_path is not None and unzip is True,
             "`unzip_file_path` must be None when `unzip` is True.")

    if unzip:
        unzip_file_fp = _load_zip_file(path)
        unzip_file_fp = str(Path(unzip_file_fp))
    else:
        unzip_file_fp = unzip_file_path

    raise_if_not(ValueError, os.listdir(unzip_file_fp).count('pipeline.pkl') == 1,
                 "Zip file must contain one file which be named `pipeline.pkl`")

    pipeline = None
    models = []
    for i in os.listdir(unzip_file_fp):
        if i == 'pipeline.pkl':
            with open(str(Path(unzip_file_fp).joinpath(i)), 'rb') as f:
                pipeline = cloudpickle.load(f)
        else:
            models.append([i[:-4], _load_single_model(str(Path(unzip_file_fp).joinpath(i)))])

    pipeline.models_ = models
    for (sub_model_name, sub_model) in pipeline.models_:
        if sub_model_name == pipeline.leader_board_.iloc[0, :]['model']:
            pipeline.best_model_ = sub_model

    shutil.rmtree(unzip_file_fp)

    return pipeline


def save_model(path, model, scaler=None):
    """
    Save a machine learning model or a pipeline to a zip file.

    [Note that]: A loaded model cannot be saved.

    Parameters
    ----------
    path : str
        The path to the zip file.

    model : object
        The fitted machine learning model or pipeline.

    scaler : object, optional
        The scaler associated with the model.

    Returns
    -------
    str
        The path to the saved zip file.
    """
    from PipelineTS.pipeline import ModelPipeline
    if isinstance(model, ModelPipeline):
        return _save_pipeline(path, model)
    else:
        return _save_single_model(path, model, scaler)


def load_model(path):
    """
    Load a machine learning model or a pipeline from a zip file.

    Parameters
    ----------
    path : str
        The path to the zip file.

    Returns
    -------
    object
        The loaded machine learning model or pipeline.
    """
    import os
    unzip_file_fp = _load_zip_file(path)

    if 'pipeline.pkl' in os.listdir(unzip_file_fp):
        return _load_pipeline(path, unzip_file_path=unzip_file_fp, unzip=False)
    else:
        return _load_single_model(path, unzip_file_path=unzip_file_fp, unzip=False)



from copy import deepcopy
import numpy as np
import pandas as pd
import re
from lightgbm import LGBMRegressor
from sklearn.preprocessing import MinMaxScaler
from spinesTS.ml_model import GBRTPreprocessing
from sklearn.multioutput import RegressorChain
from spinesUtils import generate_function_kwargs, ParameterValuesAssert
from spinesUtils.asserts import raise_if_not, raise_if
from spinesUtils.preprocessing import gc_collector, reshape_if

from PipelineTS.base.base import GBDTModelMixin, IntervalEstimationMixin
from PipelineTS.base.spines_base import SpinesMLModelMixin
from PipelineTS.utils import update_dict_without_conflict, check_time_col_is_timestamp


class WideGBRTModel(GBDTModelMixin, IntervalEstimationMixin, SpinesMLModelMixin):
    def __init__(
            self,
            time_col,
            target_col,
            lags=1,
            n_estimators=100,
            quantile=0.9,
            random_state=None,
            differential_n=1,
            moving_avg_n=0,
            extend_daily_target_features=True,
            estimator=LGBMRegressor,
            **model_init_configs
    ):
        """
        Wide Gradient Boosting Regression Trees (GBRT) Model.

        Parameters
        ----------
        time_col : str
            The column containing time information in the input data.
        target_col : str
            The column containing the target variable in the input data.
        lags : int, optional, default: 1
            The number of lagged values to use as input features for training and prediction.
        n_estimators : int, optional, default: 100
            The number of boosting rounds (trees) in the GBRT model.
        quantile : float, optional, default: 0.9
            The quantile used for interval prediction. Set to None for point prediction.
        random_state : int or None, optional, default: None
            The random seed for reproducibility.
        differential_n : int, optional, default: 0
            The number of differencing operations to apply to the target variable.
        moving_avg_n : int, optional, default: 0
            The window size for the moving average operation on the target variable.
        extend_daily_target_features : bool, optional, default: True
            Whether to extend the features with daily target-related features.
        estimator : sklearn.base.BaseEstimator, optional, default: LGBMRegressor
            The base estimator used for the GBRT model.
        **model_init_configs : dict
            Additional keyword arguments for configuring the base estimator.

        Attributes
        ----------
        estimator : BaseEstimator
            The base estimator for the GBRT model.
        processor : spinesTS.ml_model.GBRTPreprocessing
            The preprocessor for transforming input data.
        model : sklearn.multioutput.RegressorChain
            The GBRT model wrapped in a regressor chain.
        x : pandas.DataFrame or None
            The last input features used for training.
        """
        super().__init__(time_col=time_col, target_col=target_col)

        self.all_configs['model_configs'] = generate_function_kwargs(
            estimator,
            n_estimators=n_estimators,
            random_state=random_state,
            **model_init_configs
        )

        if 'LGBMRegressor' in re.split("<|>|class| |\.|\'", str(estimator)):
            self.all_configs['model_configs'] = update_dict_without_conflict(self.all_configs['model_configs'], {
                'verbose': -1
            })

        self._estimator = estimator

        self.last_dt = None
        self.last_lags_dataframe = None

        self.all_configs.update(
            {
                'lags': lags,
                'quantile': quantile,
                'time_col': time_col,
                'target_col': target_col,
                'quantile_error': 0,
                'differential_n': differential_n,
                'moving_avg_n': moving_avg_n,
                'extend_daily_target_features': extend_daily_target_features,
                'built_in_scaler': MinMaxScaler()
            }
        )

        self.processor = GBRTPreprocessing(
            in_features=self.all_configs['lags'],
            out_features=self.all_configs['lags'],
            target_col=self.all_configs['target_col'],
            date_col=self.all_configs['time_col'],
            differential_n=self.all_configs['differential_n'],
            moving_avg_n=self.all_configs['moving_avg_n'],
            extend_daily_target_features=self.all_configs['extend_daily_target_features'],
            train_size=None
        )

        self.x = None

        self.model = self._define_model()

    def _define_model(self):
        """
        Define the GBRT model using a regressor chain.

        Returns
        -------
        sklearn.multioutput.RegressorChain
            The GBRT model wrapped in a regressor chain.
        """
        return RegressorChain(self._estimator(**self.all_configs['model_configs']))

    @ParameterValuesAssert({
        'mode': ('train', 'predict')
    })
    def _data_preprocess(self, data, mode='train', update_last_dt=False):
        """
        Preprocess the input data for training or prediction.

        Parameters
        ----------
        data : pandas.DataFrame
            The input data.
        mode : {'train', 'predict'}, optional, default: 'train'
            The mode indicating whether to preprocess data for training or prediction.
        update_last_dt : bool, optional, default: False
            Whether to update the last timestamp attribute.

        Returns
        -------
        tuple or numpy.ndarray
            If 'mode' is 'train', returns a tuple (x_train, y_train).
            If 'mode' is 'predict', returns lagged splits of the target column.
        """
        data[self.all_configs['time_col']] = pd.to_datetime(data[self.all_configs['time_col']])
        if update_last_dt:
            self.last_dt = data[self.all_configs['time_col']].max()

        if mode == 'train':
            self.processor.fit(data)

            x, y = self.processor.transform(data, mode=mode)  # X, y
            return x.astype(np.float32), y.astype(np.float32)
        else:
            # x
            return self.processor.transform(data, mode=mode).astype(np.float32)

    @gc_collector()
    def fit(self, data, cv=5, fit_kwargs=None):
        """
        Fit the GBRT model to the training data.

        Parameters
        ----------
        data : pandas.DataFrame
            The training data.
        cv : int, optional, default: 5
            The number of cross-validation folds.
        fit_kwargs : dict or None, optional, default: None
            Additional keyword arguments for the fitting process.

        Returns
        -------
        self
            Returns an instance of the fitted model.
        """
        check_time_col_is_timestamp(data, self.all_configs['time_col'])

        data = data[[self.all_configs['time_col'], self.all_configs['target_col']]]

        self.last_lags_dataframe = data.iloc[-self.all_configs['lags']:, :]

        if fit_kwargs is None:
            fit_kwargs = {}

        x, y = self._data_preprocess(data, 'train', update_last_dt=True)
        x = pd.DataFrame(x)

        self.x = pd.DataFrame(
            self._data_preprocess(data, 'predict')
        ).iloc[-1:, :]

        self.model.fit(x, y, **fit_kwargs)

        if self.all_configs['quantile'] is not None:
            self.all_configs['quantile_error'] = \
                self.calculate_confidence_interval_gbrt(data, fit_kwargs=fit_kwargs, cv=cv)

        del x, y

        return self

    def _extend_predict(self, x, n):
        """
        Extrapolation prediction.

        Parameters
        ----------
        x : numpy.ndarray
            To-predict data with 2 dimensions.
        n : int
            Number of prediction steps.

        Returns
        -------
        list
            List of predicted values.
        """
        assert isinstance(n, int)
        assert x.ndim == 2

        current_res = self.model.predict(x)  # np.ndarray
        current_res = reshape_if(current_res, current_res.ndim == 1, (1, -1))

        if n <= current_res.shape[1]:
            return current_res.squeeze().tolist()[:n]
        else:
            res = current_res.squeeze().tolist()
            last_data = self.last_lags_dataframe.copy()

            last_data[self.all_configs['time_col']] = pd.to_datetime(last_data[self.all_configs['time_col']])

            last_dt = deepcopy(self.last_dt)
            for i in range(n - self.all_configs['lags']):
                tmp_data = pd.DataFrame(columns=[self.all_configs['time_col'], self.all_configs['target_col']])
                tmp_data[self.all_configs['time_col']] = (last_dt +
                                                          pd.to_timedelta(range(self.all_configs['lags'] + 1),
                                                                          unit='D'))[1:]

                tmp_data[self.all_configs['target_col']] = res[-self.all_configs['lags']:]
                last_data = pd.concat((last_data.iloc[1:, :], tmp_data.iloc[:1, :]), axis=0)

                last_dt = last_data[self.all_configs['time_col']].max()

                to_predict_x = pd.DataFrame(
                    self._data_preprocess(last_data, 'predict')
                ).iloc[-1:, :]

                current_res = self.model.predict(to_predict_x).squeeze()
                res.append(current_res[0])

            return res

    def predict(self, n, data=None):
        """
        Predict future values using the trained GBRT model.

        Parameters
        ----------
        n : int
            Number of steps to predict into the future.
        data : pandas.DataFrame or None, optional, default: None
            Additional data for prediction. If provided, the length of the series must be greater than or equal to lags.

        Returns
        -------
        pandas.DataFrame
            DataFrame containing the predicted values and corresponding timestamps.
        """
        if data is not None:
            check_time_col_is_timestamp(data, self.all_configs['time_col'])

            raise_if_not(
                ValueError, len(data) >= self.all_configs['lags'],
                'The length of the series must be greater than or equal to the lags.'
            )

            x = self._data_preprocess(data.iloc[-self.all_configs['lags']:, :], 'predict', update_last_dt=False)
            last_dt = data[self.all_configs['time_col']].max()
        else:
            x = self.x.values
            last_dt = self.last_dt

        res = self._extend_predict(x, n)  # list

        assert len(res) == n
        res = pd.DataFrame(res, columns=[self.all_configs['target_col']])
        res[self.all_configs['time_col']] = \
            last_dt + pd.to_timedelta(range(res.index.shape[0] + 1), unit='D')[1:]

        if self.all_configs['quantile'] is not None:
            res = self.interval_predict(res)

        return self.chosen_cols(res)



import logging

import torch
from darts.models.forecasting.transformer_model import TransformerModel as tfm
from spinesUtils.asserts import generate_function_kwargs

from PipelineTS.base.base import NNModelMixin, IntervalEstimationMixin
from PipelineTS.base.darts_base import DartsForecastMixin
from PipelineTS.utils import update_dict_without_conflict

logging.getLogger("pytorch_lightning.utilities.rank_zero").setLevel(logging.WARNING)
logging.getLogger("pytorch_lightning.accelerators.cuda").setLevel(logging.WARNING)


class TransformerModel(DartsForecastMixin, NNModelMixin, IntervalEstimationMixin):
    def __init__(
            self,
            time_col,
            target_col,
            lags=6,
            d_model=64,
            nhead=4,
            num_encoder_layers=3,
            num_decoder_layers=3,
            dim_feedforward=512,
            dropout=0.1,
            activation='relu',
            norm_type=None,
            custom_encoder=None,
            custom_decoder=None,
            loss_fn=torch.nn.MSELoss(),
            torch_metrics=None,
            optimizer_cls=torch.optim.Adam,
            optimizer_kwargs=None,
            lr_scheduler_cls=None,
            lr_scheduler_kwargs=None,
            use_reversible_instance_norm=False,
            batch_size=32,
            n_epochs=100,
            nr_epochs_val_period=1,
            add_encoders=None,
            enable_progress_bar=False,
            enable_model_summary=False,
            pl_trainer_kwargs=None,
            quantile=0.9,
            random_state=None,
            accelerator=None
    ):
        """
        TransformerModel: A wrapper for the TransformerModel from the darts library with additional features.

        Parameters
        ----------
        time_col : str
            The column containing time information in the input data.
        target_col : str
            The column containing the target variable in the input data.
        lags : int, optional, default: 6
            The number of lagged values to use as input features for training and prediction.
        d_model : int, optional, default: 64
            Dimensionality of the model.
        nhead : int, optional, default: 4
            Number of heads in the multiheadattention models.
        num_encoder_layers : int, optional, default: 3
            Number of sub-encoder-layers in the encoder.
        num_decoder_layers : int, optional, default: 3
            Number of sub-decoder-layers in the decoder.
        dim_feedforward : int, optional, default: 512
            Dimension of the feedforward network model.
        dropout : float, optional, default: 0.1
            The dropout value for training.
        activation : str, optional, default: 'relu'
            The activation function to use. Default is 'relu'.
        norm_type : str, optional, default: None
            Normalization type to apply on the input data. Default is None.
        custom_encoder : None, optional, default: None
            Custom encoder module to replace the default transformer encoder.
        custom_decoder : None, optional, default: None
            Custom decoder module to replace the default transformer decoder.
        loss_fn : torch.nn.modules.loss._Loss, optional, default: torch.nn.MSELoss()
            Loss function for training the model.
        torch_metrics : None, optional, default: None
            Additional metrics to track during training.
        optimizer_cls : torch.optim.Optimizer, optional, default: torch.optim.Adam
            Optimizer class for training the model.
        optimizer_kwargs : None, optional, default: None
            Additional keyword arguments for the optimizer.
        lr_scheduler_cls : None, optional, default: None
            Learning rate scheduler class for adjusting the learning rate during training.
        lr_scheduler_kwargs : None, optional, default: None
            Additional keyword arguments for the learning rate scheduler.
        use_reversible_instance_norm : bool, optional, default: False
            Whether to use reversible instance normalization.
        batch_size : int, optional, default: 32
            Batch size used during training.
        n_epochs : int, optional, default: 100
            Number of epochs for training.
        nr_epochs_val_period : int, optional, default: 1
            Number of epochs per validation period.
        add_encoders : None, optional, default: None
            Additional encoders to concatenate with the input data.
        enable_progress_bar : bool, optional, default: False
            Whether to display a progress bar during training.
        enable_model_summary : bool, optional, default: False
            Whether to display the model summary.
        pl_trainer_kwargs : None, optional, default: None
            Additional keyword arguments for the PyTorch Lightning trainer.
        quantile : float, optional, default: 0.9
            The quantile used for interval prediction. Set to None for point prediction.
        random_state : int or None, optional, default: None
            The random seed for reproducibility.
        accelerator : str, optional, default: None
            The PyTorch Lightning accelerator to use during training.

        Attributes
        ----------
        x : None
            Placeholder for input data (not used in this implementation).
        model : darts.models.forecasting.transformer_model.TransformerModel
            The TransformerModel model from the darts library.
        """
        super().__init__(time_col=time_col, target_col=target_col, accelerator=accelerator)

        if pl_trainer_kwargs is None:
            pl_trainer_kwargs = {}

        pl_trainer_kwargs = update_dict_without_conflict(pl_trainer_kwargs, {
            'accelerator': self.accelerator,
            'enable_progress_bar': enable_progress_bar,
            'enable_model_summary': enable_model_summary
        })

        self.all_configs['model_configs'] = generate_function_kwargs(
            tfm,
            input_chunk_length=lags,
            output_chunk_length=lags,
            d_model=d_model,
            nhead=nhead,
            num_encoder_layers=num_encoder_layers,
            num_decoder_layers=num_decoder_layers,
            dim_feedforward=dim_feedforward,
            dropout=dropout,
            activation=activation,
            norm_type=norm_type,
            custom_encoder=custom_encoder,
            custom_decoder=custom_decoder,
            loss_fn=loss_fn,
            torch_metrics=torch_metrics,
            optimizer_cls=optimizer_cls,
            optimizer_kwargs=optimizer_kwargs,
            lr_scheduler_cls=lr_scheduler_cls,
            lr_scheduler_kwargs=lr_scheduler_kwargs,
            use_reversible_instance_norm=use_reversible_instance_norm,
            batch_size=batch_size,
            n_epochs=n_epochs,
            nr_epochs_val_period=nr_epochs_val_period,
            add_encoders=add_encoders,
            pl_trainer_kwargs=pl_trainer_kwargs,
            random_state=random_state,
        )
        self.model = self._define_model()

        self.all_configs.update(
            {
                'lags': lags,
                'quantile': quantile,
                'time_col': time_col,
                'target_col': target_col,
                'enable_progress_bar': enable_progress_bar,
                'enable_model_summary': enable_model_summary
            }
        )

    def _define_model(self):
        """
        Define the TransformerModel model from the darts library.

        Returns
        -------
        darts.models.forecasting.transformer_model.TransformerModel
            The TransformerModel model from the darts library.
        """
        return tfm(**self.all_configs['model_configs'])



import logging

import torch
from darts.models.forecasting.tide_model import TiDEModel as TiDE
from spinesUtils.asserts import generate_function_kwargs

from PipelineTS.base.base import NNModelMixin, IntervalEstimationMixin
from PipelineTS.base.darts_base import DartsForecastMixin
from PipelineTS.utils import update_dict_without_conflict

logging.getLogger("pytorch_lightning.utilities.rank_zero").setLevel(logging.WARNING)
logging.getLogger("pytorch_lightning.accelerators.cuda").setLevel(logging.WARNING)


class TiDEModel(DartsForecastMixin, NNModelMixin, IntervalEstimationMixin):
    def __init__(
            self,
            time_col,
            target_col,
            lags=6,
            num_encoder_layers=1,
            num_decoder_layers=1,
            decoder_output_dim=16,
            hidden_size=128,
            temporal_width_past=4,
            temporal_width_future=4,
            temporal_decoder_hidden=32,
            use_layer_norm=False,
            dropout=0.1,
            use_static_covariates=True,
            loss_fn=torch.nn.MSELoss(),
            torch_metrics=None,
            optimizer_cls=torch.optim.Adam,
            optimizer_kwargs=None,
            lr_scheduler_cls=None,
            lr_scheduler_kwargs=None,
            use_reversible_instance_norm=False,
            batch_size=32,
            n_epochs=100,
            nr_epochs_val_period=1,
            add_encoders=None,
            enable_progress_bar=False,
            enable_model_summary=False,
            pl_trainer_kwargs=None,
            quantile=0.9,
            random_state=None,
            accelerator=None
    ):
        """
        TiDEModel: A wrapper for the TiDE model from the darts library with additional features.

        Parameters
        ----------
        time_col : str
            The column containing time information in the input data.
        target_col : str
            The column containing the target variable in the input data.
        lags : int, optional, default: 6
            The number of lagged values to use as input features for training and prediction.
        num_encoder_layers : int, optional, default: 1
            The number of encoder layers in the TiDE model.
        num_decoder_layers : int, optional, default: 1
            The number of decoder layers in the TiDE model.
        decoder_output_dim : int, optional, default: 16
            The output dimension of the decoder in the TiDE model.
        hidden_size : int, optional, default: 128
            The size of the hidden layer in the TiDE model.
        temporal_width_past : int, optional, default: 4
            The temporal width for past information in the TiDE model.
        temporal_width_future : int, optional, default: 4
            The temporal width for future information in the TiDE model.
        temporal_decoder_hidden : int, optional, default: 32
            The size of the hidden layer in the temporal decoder of the TiDE model.
        use_layer_norm : bool, optional, default: False
            Whether to use layer normalization in the TiDE model.
        dropout : float, optional, default: 0.1
            The dropout rate used during training.
        use_static_covariates : bool, optional, default: True
            Whether to use static covariates in the TiDE model.
        loss_fn : torch.nn.Module, optional, default: torch.nn.MSELoss()
            The loss function used for training the model.
        torch_metrics : list or None, optional, default: None
            The list of additional metrics to use during training. Set to None if no additional metrics are required.
        optimizer_cls : torch.optim.Optimizer, optional, default: torch.optim.Adam
            The optimizer used for training the model.
        optimizer_kwargs : dict or None, optional, default: None
            Additional keyword arguments for the optimizer. Set to None if no additional arguments are required.
        lr_scheduler_cls : torch.optim.lr_scheduler._LRScheduler, optional, default: None
            The learning rate scheduler used during training. Set to None if no scheduler is required.
        lr_scheduler_kwargs : dict or None, optional, default: None
            Additional keyword arguments for the learning rate scheduler. Set to None if no additional arguments are required.
        use_reversible_instance_norm : bool, optional, default: False
            Whether to use reversible instance normalization in the TiDE model.
        batch_size : int, optional, default: 32
            The batch size used during training.
        n_epochs : int, optional, default: 100
            The number of epochs for training the model.
        nr_epochs_val_period : int, optional, default: 1
            The frequency at which to perform validation during training.
        add_encoders : dict or None, optional, default: None
            The dictionary containing additional encoders for input features. Set to None if no additional encoders are required.
        enable_progress_bar : bool, optional, default: False
            Whether to display a progress bar during training.
        enable_model_summary : bool, optional, default: False
            Whether to display a summary of the model architecture.
        pl_trainer_kwargs : dict or None, optional, default: None
            Additional keyword arguments for the PyTorch Lightning trainer. Set to None if no additional arguments are required.
        quantile : float, optional, default: 0.9
            The quantile used for interval prediction. Set to None for point prediction.
        random_state : int or None, optional, default: None
            The random seed for reproducibility.
        accelerator : str or None, optional, default: None
            The PyTorch Lightning accelerator to use during training. Set to None for the default accelerator.

        Attributes
        ----------
        model : darts.models.forecasting.tide_model.TiDEModel
            The TiDE model from the darts library.
        """
        super().__init__(time_col=time_col, target_col=target_col, accelerator=accelerator)

        if pl_trainer_kwargs is None:
            pl_trainer_kwargs = {}

        pl_trainer_kwargs = update_dict_without_conflict(pl_trainer_kwargs, {
            'accelerator': self.accelerator,
            'enable_progress_bar': enable_progress_bar,
            'enable_model_summary': enable_model_summary
        })

        self.all_configs['model_configs'] = generate_function_kwargs(
            TiDE,
            input_chunk_length=lags,
            output_chunk_length=lags,
            num_encoder_layers=num_encoder_layers,
            num_decoder_layers=num_decoder_layers,
            decoder_output_dim=decoder_output_dim,
            hidden_size=hidden_size,
            temporal_width_past=temporal_width_past,
            temporal_width_future=temporal_width_future,
            temporal_decoder_hidden=temporal_decoder_hidden,
            use_layer_norm=use_layer_norm,
            dropout=dropout,
            use_static_covariates=use_static_covariates,
            loss_fn=loss_fn,
            torch_metrics=torch_metrics,
            optimizer_cls=optimizer_cls,
            optimizer_kwargs=optimizer_kwargs,
            lr_scheduler_cls=lr_scheduler_cls,
            lr_scheduler_kwargs=lr_scheduler_kwargs,
            use_reversible_instance_norm=use_reversible_instance_norm,
            batch_size=batch_size,
            n_epochs=n_epochs,
            nr_epochs_val_period=nr_epochs_val_period,
            add_encoders=add_encoders,
            pl_trainer_kwargs=pl_trainer_kwargs,
            random_state=random_state,
        )
        self.model = self._define_model()

        self.all_configs.update(
            {
                'lags': lags,
                'quantile': quantile,
                'time_col': time_col,
                'target_col': target_col,
                'enable_progress_bar': enable_progress_bar,
                'enable_model_summary': enable_model_summary
            }
        )

    def _define_model(self):
        """
        Define the TiDE model from the darts library.

        Returns
        -------
        darts.models.forecasting.tide_model.TiDEModel
            The TiDE model from the darts library.
        """
        return TiDE(**self.all_configs['model_configs'])



import time
from copy import deepcopy
import gc

import pandas as pd
from sklearn.preprocessing import QuantileTransformer, MinMaxScaler
from sklearn.base import TransformerMixin
from frozendict import frozendict

from spinesTS.base import detect_available_device
from spinesTS.metrics import mae
from spinesUtils.preprocessing import gc_collector
from spinesUtils.asserts import (
    ParameterTypeAssert,
    ParameterValuesAssert,
    check_obj_is_function,
    augmented_isinstance,
    raise_if,
    raise_if_not,
    check_has_param
)
from spinesUtils.logging import Logger
from spinesUtils.timer import Timer

# All model classes in PipelineTS are subclasses of the IntervalEstimationMixin class.
from PipelineTS.base.base import IntervalEstimationMixin
from PipelineTS.metrics import quantile_acc
from PipelineTS.pipeline.pipeline_models import get_all_available_models, get_all_model_class_name
from PipelineTS.pipeline.pipeline_configs import PipelineConfigs
from PipelineTS.utils import update_dict_without_conflict, check_time_col_is_timestamp
from PipelineTS.base.base_utils import generate_models_set


# TODO: 传入数据，进行数据采集周期检验，看看是否有漏数据，如果有，进行插值（可选），如果有异常值，进行噪音去除（可选）


class ModelPipeline:
    @ParameterTypeAssert({
        'time_col': str,
        'target_col': str,
        'lags': int,
        'quantile': (None, float),
        'exclude_models': (None, list, str),
        'metric_less_is_better': bool,
        'configs': (None, PipelineConfigs),
        'random_state': (int, None),
        'include_init_config_model': bool,
        'accelerator': (str, None),
        'cv': int,
        'gbdt_differential_n': int
    }, 'Pipeline')
    @ParameterValuesAssert({
        'metric': lambda s: check_obj_is_function(s),
        'scaler': lambda s: augmented_isinstance(s, (TransformerMixin, None, bool)),
        'accelerator': (
                lambda s:
                s in ("cpu", "gpu", "tpu", "ipu", "hpu", "mps", "auto", "cuda")
                or augmented_isinstance(s, None)
        ),
        'include_models': (
                lambda s:
                s in ('light', 'all', 'nn', 'ml') or
                augmented_isinstance(s, (None, list, str)) or
                issubclass(s, IntervalEstimationMixin)
        )
    }, 'Pipeline')
    def __init__(
            self,
            time_col,
            target_col,
            lags,
            quantile=None,  # the quantile prediction switch
            include_models='light',
            exclude_models=None,
            metric=mae,
            metric_less_is_better=True,
            configs=None,
            random_state=0,
            include_init_config_model=False,
            scaler=None,  # False for MinMaxScaler, True for StandardScaler, None means no data be scaled
            accelerator='auto',
            cv=5,
            gbdt_differential_n=1,
            **model_init_kwargs
    ):
        """
        Initialize the ModelPipeline.

        Parameters
        ----------
        time_col : str
            Name of the column representing time.
        target_col : str
            Name of the column containing the target variable.
        lags : int
            Number of lagged time steps for modeling.
        quantile : float, optional, default: None
            Quantile value of interval prediction.
        include_models : {'light', 'all', 'nn', 'ml'} or list or None or a available model of PipelineTS, optional, default: 'light'
            Models to include in the pipeline.
        exclude_models : list or None or str, optional, default: None
            Models to exclude from the pipeline.
        metric : callable, optional, default: Mean Absolute Error (mae)
            Evaluation metric function.
        metric_less_is_better : bool, optional, default: True
            Whether lower metric values are better.
        configs : PipelineConfigs or None, optional, default: None
            Configuration object for the pipeline.
        random_state : int, optional, default: 0
            Seed for random number generation.
        verbose : bool or int, optional, default: True
            Verbosity level.
        include_init_config_model : bool, optional, default: False
            Include models with initial configuration.
        scaler : bool or None or transformer that has the type of sklearn.base.TransformerMixin, optional, default: False
            Use scaler for data scaling, False for MinMaxScaler, True for QuantileTransformer, None means no scaling.
            Alternatively, you can specify your own transformer.
        accelerator : {'cpu', 'gpu', 'tpu', 'ipu', 'hpu', 'mps', 'auto', 'cuda'} or None, optional, default: 'auto'
            Hardware accelerator type.
        cv : int, optional, default: 5
            Number of cross-validation folds.
        gbdt_differential_n : int, optional, default: 1
            The number of differencing operations to apply to the target variable.
        **model_init_kwargs
            Additional keyword arguments for model initialization.

        Raises
        ------
        ValueError
            If include_models and exclude_models are set simultaneously.
            If quantile is not None and cv is not greater than 1.
            If exclude_models contain invalid model names.
            If include_models contain invalid model names.
            If model names in model_init_kwargs do not match available models.

        Notes
        -----
        The include_models parameter supports predefined sets ('light', 'all', 'nn', 'ml') or a custom list of model names.
        The accelerator parameter supports values ('cpu', 'gpu', 'tpu', 'ipu', 'hpu', 'mps', 'auto', 'cuda') or None.
        """
        raise_if(ValueError, include_models is not None and exclude_models is not None,
                 "include_models and exclude_models can not be set at the same time.")

        if augmented_isinstance(exclude_models, str):
            exclude_models = [exclude_models]

        if include_models == 'light':
            include_models = ['d_linear', 'lightgbm', 'multi_step_model', 'n_hits', 'n_linear',
                              'random_forest', 'regressor_chain', 'tcn', 'xgboost']
        elif include_models == 'all':
            include_models = None
        elif include_models == 'nn':
            include_models = ['d_linear', 'gau', 'n_beats', 'n_hits', 'n_linear', 'tcn', 'tft',
                              'patch_rnn', 'stacking_rnn', 'tide', 'time2vec', 'transformer']
        elif include_models == 'ml':
            include_models = ['catboost', 'lightgbm', 'multi_output_model',
                              'multi_step_model', 'random_forest', 'wide_gbrt', 'xgboost']
        elif isinstance(include_models, str):
            raise_if_not(ValueError, include_models in ModelPipeline.list_all_available_models(),
                         f"{include_models} is not a available model name. ")
            include_models = [include_models]
        elif not isinstance(include_models, (list, str)) and issubclass(include_models, IntervalEstimationMixin):
            include_models = [include_models]
        else:
            include_models = include_models

        if quantile:
            raise_if(ValueError, cv <= 1, "if quantile is not None, cv must be greater than 1.")

        self._available_models = get_all_available_models()

        raise_if(ValueError, exclude_models is not None and
                 (not all([i in self._available_models for i in exclude_models])),
                 "exclude_models must be None or in the list of models.")

        raise_if(ValueError, include_models is not None and
                 (not all([i in self._available_models or
                           issubclass(i, IntervalEstimationMixin) for i in include_models])),
                 "include_models must be None or in the list of models or a available PipelineTS model.")

        if exclude_models is not None:
            self._available_models = dict(self._available_models)

            for em in exclude_models:
                del self._available_models[em]
            self._available_models = frozendict(self._available_models)

        self.logger = Logger(name='PipelineTS')

        self.target_col = target_col
        self.time_col = time_col
        self.lags = lags
        self.metric = metric
        self.metric_less_is_better = metric_less_is_better
        self.random_state = random_state
        self.configs = configs
        self._given_models = include_models
        self.quantile = quantile

        self.include_init_config_model = include_init_config_model

        if augmented_isinstance(scaler, bool):
            self.scaler = QuantileTransformer() if scaler else MinMaxScaler()
        else:
            self.scaler = scaler

        self._temp_scaler = deepcopy(self.scaler)

        self.models_ = []
        self.leader_board_ = None
        self.best_model_ = None
        self.accelerator = accelerator
        self.cv = cv

        self._timer = Timer()

        self._model_init_kwargs = {}

        model_init_kwargs = update_dict_without_conflict(model_init_kwargs,
                                                         {
                                                             'multi_output_model__verbose': -1,
                                                             'multi_step_model__verbose': -1,
                                                             'lightgbm__verbose': -1,
                                                             'wide_gbrt__verbose': -1,
                                                             'catboost__verbose': False,
                                                             'xgboost__verbose': 0
                                                         })

        for k, v in model_init_kwargs.items():
            raise_if(ValueError, '__' not in k,
                     f"{k} must has double underline.")

            raise_if(ValueError, k.split('__')[0] not in self._available_models and k.split('__')[0]
                     not in get_all_model_class_name(),
                     f"{k.split('__')[0]} is not a valid model name")
            self._model_init_kwargs[k] = v

        self._compute_device_msg = detect_available_device(self.accelerator)[1] + '\n\n'

        self.gbdt_differential_n = gbdt_differential_n

    def _initial_models(self):
        initial_models = []
        ms = generate_models_set(self._available_models, self._given_models)

        # 模型训练顺序
        for (model_name, model) in ms:
            model_kwargs = self._fill_func_params(
                func=model,
                time_col=self.time_col,
                target_col=self.target_col,
                lags=self.lags,
                random_state=self.random_state,
                quantile=self.quantile,
                accelerator=self.accelerator,
                differential_n=self.gbdt_differential_n
            )

            # Populate model initialization parameters specified in double underscore format.
            # This scenario takes precedence over keyword arguments.
            if len(self._model_init_kwargs) > 0:
                for k, v in self._model_init_kwargs.items():
                    if k.split('__')[0] == model_name:
                        model_kwargs[k[len(model_name) + 2:]] = v

            # The PipelineConfigs class has the highest configuration authority.
            if self.configs is not None:
                include_in_configs = False
                for (model_name_in_config, model_name_after_rename_in_config, model_configs_in_config) \
                        in self.configs.configs:
                    if model_name_in_config == model_name:
                        include_in_configs = True
                        new_model_kwargs = deepcopy(model_kwargs)

                        new_model_kwargs.update(
                            self.configs.get_configs(model_name_after_rename_in_config).get('init_configs')
                        )

                        initial_models.append([model_name_after_rename_in_config, model(**new_model_kwargs)])

                        if self.include_init_config_model:
                            if [model_name, model(**model_kwargs)] not in initial_models:
                                initial_models.append([model_name, model(**model_kwargs)])

                if not include_in_configs:
                    initial_models.append([model_name, model(**model_kwargs)])
            else:
                initial_models.append([model_name, model(**model_kwargs)])

        return initial_models

    @classmethod
    def list_all_available_models(cls):
        """
        Get a list of all available model names in the ModelPipeline.

        Returns
        -------
        models : list of str
            List of model names available for use in the pipeline.

        Example
        -------
        >>> ModelPipeline.list_all_available_models()
        ['catboost',
         'd_linear',
         'gau',
         'lightgbm',
         'multi_output_model',
         'multi_step_model',
         'n_beats',
         'n_hits',
         'n_linear',
         'random_forest',
         'regressor_chain',
         'patch_rnn',
         'stacking_rnn',
         'tcn',
         'tft',
         'tide',
         'time2vec',
         'transformer',
         'xgboost']
        """
        return sorted(list(get_all_available_models().keys()))

    def _scale_data(self, data, valid_data=None, refit_scaler=True):
        df, valid_df = data.copy(), valid_data  # valid_data will not be deep copy in this step

        if refit_scaler:
            scaler = self.scaler
        else:
            scaler = self._temp_scaler

        if scaler is not None:
            df[self.target_col] = scaler.fit_transform(
                df[self.target_col].values.reshape(-1, 1)
            ).squeeze()

            if valid_data is not None:
                valid_df = valid_data.copy()
                valid_df[self.target_col] = scaler.transform(
                    valid_df[self.target_col].values.reshape(-1, 1)).squeeze()

        return df, valid_df

    def _inverse_data(self, df, columns=None, use_scaler=True):
        if use_scaler:
            scaler = self.scaler
        else:
            scaler = self._temp_scaler

        if columns is None:
            columns = self.target_col

        if scaler is not None:
            df[columns] = scaler.inverse_transform(
                df[columns].values.reshape(-1, 1)
            ).squeeze()

        return df

    @staticmethod
    def _fill_func_params(func, **kwargs):
        init_kwargs = {}

        for i in kwargs:
            if check_has_param(func, i):
                init_kwargs.update({i: kwargs[i]})

        return init_kwargs

    @gc_collector(3)
    def _fit(self, model_name_after_rename, model, train_df, valid_df, res_df, use_scaler=True):
        self._timer.start()

        # -------------------- fitting -------------------------
        if self.configs is not None:
            if self.configs.get_configs(model_name_after_rename):
                fit_kwargs = self.configs.get_configs(model_name_after_rename).get('fit_configs')
            else:
                fit_kwargs = {}
        else:
            fit_kwargs = {}

        model_kwargs = self._fill_func_params(func=model.fit, data=train_df, fit_kwargs=fit_kwargs, cv=self.cv,
                                              valid_data=valid_df)
        model.fit(**model_kwargs)

        train_cost = self._timer.last_timestamp_diff()

        self._timer.middle_point()
        gc.collect()
        gc.garbage.clear()

        self._timer.sleep(3)
        # -------------------- predicting -------------------------
        if self.configs is not None:
            if self.configs.get_configs(model_name_after_rename):
                predict_kwargs = self.configs.get_configs(model_name_after_rename).get('predict_configs')
            else:
                predict_kwargs = {}
        else:
            predict_kwargs = {}

        if check_has_param(model.predict, 'predict_kwargs'):
            eval_res = model.predict(valid_df.shape[0], data=valid_df, predict_kwargs=predict_kwargs)
        else:
            eval_res = model.predict(valid_df.shape[0])

        if use_scaler:
            scaler = self.scaler
        else:
            scaler = self._temp_scaler

        yt = valid_df[self.target_col].values
        yp = eval_res[self.target_col].values

        if self.quantile:
            left_pred = eval_res[f"{self.target_col}_lower"].values
            right_pred = eval_res[f"{self.target_col}_upper"].values

        if scaler is not None:
            yt = scaler.inverse_transform(yt.reshape(-1, 1)).squeeze()
            yp = scaler.inverse_transform(yp.reshape(-1, 1)).squeeze()

            if self.quantile:
                left_pred = scaler.inverse_transform(left_pred.reshape(-1, 1)).squeeze()
                right_pred = scaler.inverse_transform(right_pred.reshape(-1, 1)).squeeze()
                res_quantile_acc = quantile_acc(yt, left_pred, right_pred)

        metric = self.metric(yt, yp)

        eval_cost = self._timer.last_timestamp_diff()

        del eval_res

        gc.collect()
        self._timer.sleep(3)

        self._timer.clear()  # 重置计时器

        if self.quantile:
            res_df = pd.concat(
                (res_df, pd.DataFrame(
                    [[model_name_after_rename, train_cost, eval_cost, metric, res_quantile_acc]],
                    columns=['model', 'train_cost(s)', 'eval_cost(s)', 'metric', 'quantile_acc'])),
                axis=0, ignore_index=True)
        else:
            res_df = pd.concat(
                (res_df, pd.DataFrame([[model_name_after_rename, train_cost, eval_cost, metric]],
                                      columns=['model', 'train_cost(s)', 'eval_cost(s)', 'metric'])),
                axis=0, ignore_index=True)

        return model_name_after_rename, model, res_df

    @ParameterTypeAssert({
        'data': pd.DataFrame,
        'valid_data': (None, pd.DataFrame)
    })
    def fit(self, data, valid_data=None):
        """
        Fit all models in the ModelPipeline to the provided training data.

        Parameters
        ----------
        data : pd.DataFrame
            The training data containing historical information.
        valid_data : pd.DataFrame or None, optional, default: None
            Validation data for evaluating model performance.

        Returns
        -------
        leaderboard : pd.DataFrame
            Leaderboard containing model evaluation metrics, sorted by model performance.

        Raises
        ------
        ValueError
            If the length of data is less than or equal to lags.
        AssertionError
            If columns of data and valid_data do not match.

        Example
        -------
        >>> pipeline = ModelPipeline(time_col='timestamp', target_col='value', lags=10)
        >>> leaderboard = pipeline.fit(train_data, valid_data)
        >>> print(leaderboard)
           Leaderboard         model  train_cost(s)  eval_cost(s)    metric
        0           0    lightgbm_0       2.567801      0.978624  0.123456
        1           1    xgboost_1       3.123456      1.234567  0.456789
        2           2  random_forest       1.987654      0.876543  0.987654
        ...         ...            ...            ...           ...       ...

        Notes
        -----
        - The fit function trains all models in the pipeline using the provided training data.
        - The optional valid_data parameter allows for model evaluation on a separate validation dataset.
        - The resulting leaderboard provides a ranked list of models based on the specified evaluation metric.
        """
        self.logger.info('Information about the device used for computation:\n'+self._compute_device_msg)
        time.sleep(0.5)

        check_time_col_is_timestamp(data, self.time_col)

        if data.shape[0] <= self.lags:
            raise ValueError(f'length of df must be greater than lags, df length = {data.shape[0]}, lags = {self.lags}')

        if valid_data is not None:
            assert data.columns.tolist() == valid_data.columns.tolist()
            check_time_col_is_timestamp(valid_data, self.time_col)

            df, valid_df = data.copy(), valid_data.copy()
        else:
            df, valid_df = data.copy(), data.iloc[-(2 * self.lags):, :]

        # 如果指定use_standard_scale，此语句会对数据缩放
        df, valid_df = self._scale_data(df, valid_df, refit_scaler=True)

        res = pd.DataFrame(columns=['model', 'train_cost(s)', 'eval_cost(s)', 'metric'])
        if self.quantile:
            res = pd.DataFrame(columns=['model', 'train_cost(s)', 'eval_cost(s)', 'metric', 'quantile_acc'])

        models = self._initial_models()
        self.logger.info(f"There are a total of {len(models)} models to be trained.")

        for idx, (model_name_after_rename, model) in enumerate(models):
            self.logger.info(f"[model {idx:>{len(str(len(models)))}d}] fitting and "
                              f"evaluating {model_name_after_rename}...")

            model_name_after_rename, model, res = self._fit(
                model_name_after_rename, model, df, valid_df, res,
                use_scaler=True
            )

            self.models_.append((model_name_after_rename, model))

        self.leader_board_ = res.sort_values(
            by='metric', ascending=self.metric_less_is_better
        ).reset_index(drop=True)

        self.leader_board_.columns.name = 'Leaderboard'

        self.best_model_ = self.get_model(self.leader_board_.iloc[0, :]['model'])

        del data, valid_data, df, valid_df, res
        gc.collect()
        gc.garbage.clear()

        return self.leader_board_

    @ParameterTypeAssert({
        'model_name': (str, None)
    })
    def get_model(self, model_name=None):
        """
        Retrieve a trained model from the ModelPipeline.

        Parameters
        ----------
        model_name : str or None, optional, default: None
            Name of the model to retrieve. If None, returns the best model.

        Returns
        -------
        model : Model
            The trained model corresponding to the specified model_name. If model_name is None, returns the best model.

        Example
        -------
        >>> pipeline = ModelPipeline(time_col='timestamp', target_col='value', lags=10)
        >>> pipeline.fit(train_data, valid_data)
        >>> best_model = pipeline.get_model()
        >>> specific_model = pipeline.get_model('lightgbm_0')

        Notes
        -----
        - If model_name is not provided, the function returns the best-performing model based on the leaderboard.
        - The function allows retrieving a specific trained model by providing its unique name (e.g., 'lightgbm_0').
        """
        if model_name is None:
            return self.best_model_
        else:
            for (md_name, model) in self.models_:
                if model_name == md_name:
                    return model

    @ParameterTypeAssert({
        'model_name': (str, None)
    })
    def get_model_all_configs(self, model_name=None):
        """
        Retrieve the configuration details of a trained model from the ModelPipeline.

        Parameters
        ----------
        model_name : str or None, optional, default: None
            Name of the model to retrieve configuration details. If None, returns the configuration details of the best model.

        Returns
        -------
        configs : dict or None
            A dictionary containing the configuration details of the specified model. If model_name is None, returns the configuration details of the best model.

        Example
        -------
        >>> pipeline = ModelPipeline(time_col='timestamp', target_col='value', lags=10)
        >>> pipeline.fit(train_data, valid_data)
        >>> best_model_configs = pipeline.get_model_all_configs()
        >>> specific_model_configs = pipeline.get_model_all_configs('lightgbm_0')

        Notes
        -----
        - If model_name is not provided, the function returns the configuration details of the best-performing model.
        - The function allows retrieving configuration details for a specific trained model by providing its unique name (e.g., 'lightgbm_0').
        """
        if model_name is None:
            return self.best_model_.all_configs
        else:
            for (md_name, model) in self.models_:
                if model_name == md_name:
                    return model.all_configs

    @ParameterTypeAssert({
        'n': int,
        'data': (pd.DataFrame, None),
        'model_name': (None, str)
    })
    def predict(self, n, data=None, model_name=None):
        """
        Generate predictions using the trained models in the ModelPipeline.

        Parameters
        ----------
        n : int
            Predictive steps, indicating the number of time steps to forecast into the future.
        data : pd.DataFrame or None, optional, default: None
            The input data for making predictions. If None, the last available data in the pipeline will be used.
        model_name : str or None, optional, default: None
            Model name to use for predictions. If None, the best model will be used.

        Returns
        -------
        predictions : pd.DataFrame
            DataFrame containing the predicted values for the specified model or the best model.

        Example
        -------
        >>> pipeline = ModelPipeline(time_col='timestamp', target_col='value', lags=10)
        >>> pipeline.fit(train_data, valid_data)
        >>> predictions_best_model = pipeline.predict(n=5)
        >>> predictions_specific_model = pipeline.predict(n=5, model_name='lightgbm_0', data=test_data)

        Notes
        -----
        - The predict function generates future predictions using the trained models in the pipeline.
        - If data is not provided, the function uses the last available data in the pipeline.
        - If model_name is not provided, the function uses the best-performing model based on the leaderboard.
        """
        df = None
        if data is not None:
            df = data.copy()
            if self.scaler is not None:
                df[self.target_col] = self.scaler.transform(df[self.target_col].values.reshape(-1, 1)).squeeze()

        if model_name is not None:
            if check_has_param(self.get_model(model_name).predict, 'data'):

                res = self.get_model(model_name).predict(n, data=df)
            else:
                res = self.get_model(model_name).predict(n)
        else:
            if check_has_param(self.get_model(model_name).predict, 'data'):

                res = self.best_model_.predict(n, data=df)
            else:
                res = self.best_model_.predict(n)

        for i in res.columns:
            if i.startswith(self.target_col):
                res = self._inverse_data(res, columns=i)

        return res



from spinesTS.data import (
    DataGenerator,
    RandomEventGenerator,
    LoadElectricDataSets,
    LoadMessagesSentHourDataSets,
    LoadMessagesSentDataSets,
    LoadWebSales,
    LoadSupermarketIncoming,
    BuiltInSeriesData
)


import logging

import torch
from darts.models.forecasting.nbeats import NBEATSModel as n_beats
from spinesUtils.asserts import generate_function_kwargs

from PipelineTS.base.base import NNModelMixin, IntervalEstimationMixin
from PipelineTS.base.darts_base import DartsForecastMixin
from PipelineTS.utils import update_dict_without_conflict

logging.getLogger("pytorch_lightning.utilities.rank_zero").setLevel(logging.WARNING)
logging.getLogger("pytorch_lightning.accelerators.cuda").setLevel(logging.WARNING)


class NBeatsModel(DartsForecastMixin, NNModelMixin, IntervalEstimationMixin):
    def __init__(
            self,
            time_col,
            target_col,
            lags=6,
            generic_architecture=False,
            num_stacks=30,
            num_blocks=1,
            num_layers=4,
            layer_widths=256,
            expansion_coefficient_dim=5,
            trend_polynomial_degree=2,
            dropout=0.0,
            activation='ReLU',
            loss_fn=torch.nn.MSELoss(),
            torch_metrics=None,
            optimizer_cls=torch.optim.Adam,
            optimizer_kwargs=None,
            lr_scheduler_cls=None,
            lr_scheduler_kwargs=None,
            use_reversible_instance_norm=False,
            batch_size=32,
            n_epochs=100,
            nr_epochs_val_period=1,
            add_encoders=None,
            enable_progress_bar=False,
            enable_model_summary=False,
            pl_trainer_kwargs=None,
            quantile=0.9,
            random_state=None,
            accelerator=None
    ):
        """
        NBeatsModel: A wrapper for the NBEATSModel from the darts library with additional features.

        Parameters
        ----------
        time_col : str
            The column containing time information in the input data.
        target_col : str
            The column containing the target variable in the input data.
        lags : int, optional, default: 6
            The number of lagged values to use as input features for training and prediction.
        generic_architecture : bool, optional, default: False
            Whether to use the generic architecture (generic forecasting model).
        num_stacks : int, optional, default: 30
            The number of stacks in the architecture.
        num_blocks : int, optional, default: 1
            The number of blocks in each stack.
        num_layers : int, optional, default: 4
            The number of fully connected layers in each block.
        layer_widths : int or list, optional, default: 256
            The number of neurons in each fully connected layer. Can be an integer or a list for variable widths.
        expansion_coefficient_dim : int, optional, default: 5
            The dimension of the expansion coefficients.
        trend_polynomial_degree : int, optional, default: 2
            The degree of the polynomial used to model the trend.
        dropout : float, optional, default: 0.0
            The dropout rate applied to each fully connected layer.
        activation : str, optional, default: 'ReLU'
            The activation function used in the fully connected layers.
        loss_fn : torch.nn.Module, optional, default: torch.nn.MSELoss()
            The loss function used for training the model.
        torch_metrics : list or None, optional, default: None
            Additional metrics to track during training.
        optimizer_cls : torch.optim.Optimizer, optional, default: torch.optim.Adam
            The optimizer class used for training.
        optimizer_kwargs : dict or None, optional, default: None
            Additional keyword arguments for the optimizer.
        lr_scheduler_cls : torch.optim.lr_scheduler._LRScheduler or None, optional, default: None
            The learning rate scheduler class used for training.
        lr_scheduler_kwargs : dict or None, optional, default: None
            Additional keyword arguments for the learning rate scheduler.
        use_reversible_instance_norm : bool, optional, default: False
            Whether to use reversible instance normalization.
        batch_size : int, optional, default: 32
            The batch size used during training.
        n_epochs : int, optional, default: 100
            The number of epochs for training the model.
        nr_epochs_val_period : int, optional, default: 1
            The period for validating the model during training.
        add_encoders : dict or None, optional, default: None
            Additional encoder configurations.
        enable_progress_bar : bool, optional, default: False
            Whether to display a progress bar during training.
        enable_model_summary : bool, optional, default: False
            Whether to print the model summary.
        pl_trainer_kwargs : dict or None, optional, default: None
            Additional keyword arguments for the PyTorch Lightning trainer.
        quantile : float, optional, default: 0.9
            The quantile used for interval prediction. Set to None for point prediction.
        random_state : int or None, optional, default: None
            The random seed for reproducibility.
        accelerator : str or None, optional, default: None
            The PyTorch Lightning accelerator to use during training.

        Attributes
        ----------
        model : darts.models.forecasting.nbeats.NBEATSModel
            The NBEATSModel from the darts library.
        """
        super().__init__(time_col=time_col, target_col=target_col, accelerator=accelerator)

        if pl_trainer_kwargs is None:
            pl_trainer_kwargs = {}

        pl_trainer_kwargs = update_dict_without_conflict(pl_trainer_kwargs, {
            'accelerator': self.accelerator,
            'enable_progress_bar': enable_progress_bar,
            'enable_model_summary': enable_model_summary
        })

        self.all_configs['model_configs'] = generate_function_kwargs(
            n_beats,
            input_chunk_length=lags,
            output_chunk_length=lags,
            generic_architecture=generic_architecture,
            num_stacks=num_stacks,
            num_blocks=num_blocks,
            num_layers=num_layers,
            layer_widths=layer_widths,
            expansion_coefficient_dim=expansion_coefficient_dim,
            trend_polynomial_degree=trend_polynomial_degree,
            dropout=dropout,
            activation=activation,
            loss_fn=loss_fn,
            torch_metrics=torch_metrics,
            optimizer_cls=optimizer_cls,
            optimizer_kwargs=optimizer_kwargs,
            lr_scheduler_cls=lr_scheduler_cls,
            lr_scheduler_kwargs=lr_scheduler_kwargs,
            use_reversible_instance_norm=use_reversible_instance_norm,
            batch_size=batch_size,
            n_epochs=n_epochs,
            nr_epochs_val_period=nr_epochs_val_period,
            add_encoders=add_encoders,
            pl_trainer_kwargs=pl_trainer_kwargs,
            random_state=random_state,
        )
        self.model = self._define_model()

        self.all_configs.update(
            {
                'lags': lags,
                'quantile': quantile,
                'time_col': time_col,
                'target_col': target_col,
                'enable_progress_bar': enable_progress_bar,
                'enable_model_summary': enable_model_summary
            }
        )

    def _define_model(self):
        """
        Define the NBEATSModel from the darts library.

        Returns
        -------
        darts.models.forecasting.nbeats.NBEATSModel
            The NBEATSModel from the darts library.
        """
        return n_beats(**self.all_configs['model_configs'])



from spinesTS.nn import Time2VecNet
from spinesUtils import generate_function_kwargs

from PipelineTS.base.spines_base import SpinesNNModelMixin


class Time2VecModel(SpinesNNModelMixin):
    def __init__(
            self,
            time_col,
            target_col,
            lags=30,
            quantile=0.9,
            random_state=None,
            learning_rate=0.01,
            accelerator='auto',
            verbose=False,
            epochs=1000,
            batch_size='auto',
            patience=100,
            min_delta=0,
            lr_scheduler='CosineAnnealingLR',
            lr_scheduler_patience=10,
            lr_factor=0.7,
            restore_best_weights=True,
            loss_type='min'
    ):
        """
        Time2VecModel: A wrapper for the Time2VecNet model with additional features.

        Parameters
        ----------
        time_col : str
            The column containing time information in the input data.
        target_col : str
            The column containing the target variable in the input data.
        lags : int, optional, default: 30
            The number of lagged values to use as input features for training and prediction.
        quantile : float, optional, default: 0.9
            The quantile used for interval prediction. Set to None for point prediction.
        random_state : int or None, optional, default: None
            The random seed for reproducibility.
        learning_rate : float, optional, default: 0.01
            The learning rate for training the model.
        accelerator : str, optional, default: 'auto'
            The PyTorch Lightning accelerator to use during training.
        verbose : bool, optional, default: False
            Whether to display detailed information during training.
        epochs : int, optional, default: 1000
            The number of epochs for training the model.
        batch_size : int or 'auto', optional, default: 'auto'
            The batch size used during training. Set to 'auto' for automatic batch size determination.
        patience : int, optional, default: 100
            The patience parameter for early stopping during training.
        min_delta : float, optional, default: 0
            The minimum change in the monitored quantity to qualify as an improvement during training.
        lr_scheduler : str, optional, default: 'CosineAnnealingLR'
            The learning rate scheduler to use during training.
        lr_scheduler_patience : int, optional, default: 10
            The patience parameter for the learning rate scheduler.
        lr_factor : float, optional, default: 0.7
            The factor by which the learning rate is reduced during training.
        restore_best_weights : bool, optional, default: True
            Whether to restore the best weights during training when using early stopping.
        loss_type : str, optional, default: 'min'
            The type of loss function to use. Supported values include 'min' and 'max'.

        Attributes
        ----------
        x : None
            Placeholder for input data (not used in this implementation).
        model : spinesTS.nn.Time2VecNet
            The Time2VecNet model from the spinesTS library.
        """
        super().__init__(time_col=time_col, target_col=target_col, accelerator=accelerator)

        self.all_configs['model_configs'] = generate_function_kwargs(
            Time2VecNet,
            in_features=lags,
            out_features=lags,
            learning_rate=learning_rate,
            random_seed=random_state,
            device=self.accelerator,
            loss_fn='mae'
        )

        self.last_dt = None

        self.all_configs.update(
            {
                'lags': lags,
                'quantile': quantile,
                'time_col': time_col,
                'target_col': target_col,
                'quantile_error': 0,
                'verbose': verbose,
                'epochs': epochs,
                'batch_size': batch_size,
                'patience': patience,
                'min_delta': min_delta,
                'lr_scheduler': lr_scheduler,
                'lr_scheduler_patience': lr_scheduler_patience,
                'lr_factor': lr_factor,
                'restore_best_weights': restore_best_weights,
                'loss_type': loss_type
            }
        )

        self.x = None

        self.model = self._define_model()

    def _define_model(self):
        """
        Define the Time2VecNet model from the spinesTS library.

        Returns
        -------
        spinesTS.nn.Time2VecNet
            The Time2VecNet model from the spinesTS library.
        """
        return Time2VecNet(**self.all_configs['model_configs'])



import logging
from spinesUtils import generate_function_kwargs
from PipelineTS.base.base import StatisticModelMixin, IntervalEstimationMixin
from PipelineTS.utils import check_time_col_is_timestamp

logger = logging.getLogger('cmdstanpy')
logger.addHandler(logging.NullHandler())
logger.propagate = False
logger.setLevel(logging.CRITICAL)


class ProphetModel(StatisticModelMixin, IntervalEstimationMixin):
    """
    ProphetModel: A wrapper for the Facebook Prophet model with additional features.

    Parameters
    ----------
    time_col : str
        The column containing time information in the input data.
    target_col : str
        The column containing the target variable in the input data.
    lags : int, optional, default: 1
        The number of lagged values to use as input features for training and prediction.
    country_holidays : dict, optional, default: None
        A dictionary specifying country-specific holidays in the format {country: [list of holidays]}.
    quantile : float, optional, default: 0.9
        The quantile used for interval prediction. Set to None for point prediction.
    random_state : int, optional, default: 0
        The random seed for reproducibility.
    **prophet_configs
        Additional keyword arguments for configuring the Prophet model.

    Attributes
    ----------
    model : Prophet
        The Prophet model from the Facebook Prophet library.
    """

    from prophet import Prophet

    def __init__(
            self,
            time_col,
            target_col,
            lags=1,
            country_holidays=None,
            quantile=0.9,
            random_state=0,
            **prophet_configs
    ):
        super().__init__(time_col=time_col, target_col=target_col)

        self.all_configs['model_configs'] = generate_function_kwargs(
            ProphetModel.Prophet,
            holidays=country_holidays,
            **prophet_configs
        )

        self.model = self._define_model()

        self.all_configs.update({
            'quantile': quantile,
            'quantile_error': 0,
            'time_col': time_col,
            'target_col': target_col,
            'random_state': random_state,
            'lags': lags,
        })

    def _define_model(self):
        """
        Define the Prophet model from the Facebook Prophet library.

        Returns
        -------
        Prophet
            The Prophet model from the Facebook Prophet library.
        """
        return ProphetModel.Prophet(**self.all_configs['model_configs'])

    @staticmethod
    def _prophet_preprocessing(df, time_col, target_col):
        """
        Preprocess the input data for compatibility with the Prophet model.

        Parameters
        ----------
        df : pd.DataFrame
            The input data in pandas DataFrame format.
        time_col : str
            The column containing time information in the input data.
        target_col : str
            The column containing the target variable in the input data.

        Returns
        -------
        pd.DataFrame
            The preprocessed DataFrame compatible with the Prophet model.
        """
        df_ = df[[time_col, target_col]]
        if 'ds' != time_col or 'y' != target_col:
            df_ = df_.rename(columns={time_col: 'ds', target_col: 'y'})
        return df_

    def fit(self, data, freq='D', cv=5, fit_kwargs=None):
        """
        Fit the Prophet model on the input data.

        Parameters
        ----------
        data : pd.DataFrame
            The input data in pandas DataFrame format.
        freq : str, optional, default: 'D'
            The frequency of the time series data.
        cv : int, optional, default: 5
            The number of cross-validation folds.
        fit_kwargs : None, optional, default: None
            Additional keyword arguments for fitting the model.

        Returns
        -------
        self : ProphetModel
            Returns the instance itself.
        """
        check_time_col_is_timestamp(data, self.all_configs['time_col'])

        if fit_kwargs is None:
            fit_kwargs = {}
        data = self._prophet_preprocessing(data, self.all_configs['time_col'], self.all_configs['target_col'])
        self.model.fit(data, **fit_kwargs)

        if self.all_configs['quantile'] is not None:
            self.all_configs['quantile_error'] = \
                self.calculate_confidence_interval_prophet(data, cv=cv, freq=freq, fit_kwargs=fit_kwargs)
        return self

    def predict(self, n, freq='D', include_history=False):
        """
        Make predictions using the fitted Prophet model.

        Parameters
        ----------
        n : int
            The number of time steps to predict.
        freq : str, optional, default: 'D'
            The frequency of the time series data.
        include_history : bool, optional, default: False
            Whether to include the historical data in the predictions.

        Returns
        -------
        pd.DataFrame
            The DataFrame containing the predicted values.
        """
        res = self.model.predict(
            self.model.make_future_dataframe(
                periods=n,
                freq=freq,
                include_history=include_history,
            )
        )[['ds', 'yhat']].rename(
            columns={
                'ds': self.all_configs['time_col'],
                'yhat': self.all_configs['target_col'],
            }
        )

        if self.all_configs['quantile'] is not None:
            res = self.interval_predict(res)

        return self.chosen_cols(res)



from pathlib import Path

from setuptools import find_packages, setup


def read_requirements(path):
    return list(Path(path).read_text().splitlines())


base_reqs = read_requirements(Path('.').parent.joinpath("requirements/core.txt"))
extra_reqs = read_requirements(Path('.').parent.joinpath("requirements/extra.txt"))

all_reqs = base_reqs + extra_reqs

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='PipelineTS',
    version="0.3.11",
    description='One-stop time series analysis tool, supporting time series data preprocessing,'
                ' feature engineering, model training, model evaluation, model prediction, etc. '
                'Based on spinesTS and darts.',
    keywords='time series forecasting',
    packages=find_packages(),
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',
        'Topic :: Software Development',
        'Topic :: Scientific/Engineering',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11'
    ],
    python_requires=">=3.9",
    url='https://github.com/BirchKwok/PipelineTS',
    author='Birch Kwok',
    author_email='birchkwok@gmail.com',
    install_requires=base_reqs,
    extras_require={"all": all_reqs, "core": base_reqs},
    zip_safe=False,
)



import pandas as pd

from spinesUtils.asserts import ParameterTypeAssert, raise_if_not


@ParameterTypeAssert({
    'data1': pd.DataFrame,
    'data2': pd.DataFrame,
    'time_col': str,
    'target_col': str,
    'labels': (None, list, tuple),
    'date_fmt': str
})
def plot_data_period(data1, data2, time_col, target_col, labels=None, date_fmt='%Y-%m-%d'):
    """
    Visualize time-series data, including predictions with upper and lower bounds.

    Parameters
    ----------
    data1 : pd.DataFrame
        First dataset containing time-series data for plotting.
    data2 : pd.DataFrame
        Second dataset containing time-series data for plotting.
    time_col : str
        Column name in the dataframes representing the time information.
    target_col : str
        Column name in the dataframes representing the target variable.
    labels : None or list or tuple, optional, default: None
        Labels for the plot. If None, default labels will be used.
        For predictions with upper and lower bounds, provide labels in the order ('True Data', 'Prediction Data', 'Upper Bound', 'Lower Bound').
        For regular data, provide labels in the order ('Data', 'Test Data').
    date_fmt : str, optional, default: '%Y-%m-%d'
        Date format for the x-axis labels.

    Raises
    ------
    AssertionError
        If the starting time of data1 is after the starting time of data2.

    Notes
    -----
    - The function uses Matplotlib to create a time-series plot with optional upper and lower bounds.
    - The x-axis is formatted with the specified date format.
    - Vertical lines, shaded regions, and grid lines are added for better visualization.
    """
    import matplotlib.pyplot as plt
    import matplotlib.dates as m_dates

    # Check if the starting time of data1 is before or equal to the starting time of data2
    assert data1[time_col].iloc[0] <= data2[time_col].iloc[0], \
        'The starting time of data1 must be before data2, or equal to the starting time of data2.'

    # Check if labels are provided and in the correct format
    if f'{target_col}_upper' in data2.columns and f'{target_col}_lower' in data2.columns:
        if labels is None:
            labels = ('True Data', 'Prediction Data', 'Upper Bound', 'Lower Bound')
        else:
            raise_if_not(ValueError, len(labels) == 4,
                         "If 'target_upper' and 'target_lower' are provided, the labels must be provided as a list of "
                         "4 strings.")
    else:
        if labels is None:
            labels = ('Data', 'Test Data')
        else:
            raise_if_not(ValueError, len(labels) == 2,
                         "If 'target_upper' and 'target_lower' are not provided, the labels must be provided as a "
                         "list of 2 strings.")

    # Set x-axis as date format
    date_fmt = m_dates.DateFormatter(date_fmt)
    plt.gca().xaxis.set_major_formatter(date_fmt)

    # Plot curves
    plt.plot_date(data1[time_col], data1[target_col], label=labels[0], color='black', fmt='')
    plt.plot_date(data2[time_col], data2[target_col], label=labels[1], color='blue', fmt='')

    # Plot upper and lower bounds if available
    if f'{target_col}_upper' in data2.columns and  f'{target_col}_lower' in data2.columns:
        plt.plot_date(data2[time_col], data2[f'{target_col}_upper'], label=labels[2], color='orchid', fmt='')
        plt.plot_date(data2[time_col], data2[f'{target_col}_lower'], label=labels[3], color='violet', fmt='')
        plt.fill_between(data2[time_col], data2[f'{target_col}_upper'], data2[f'{target_col}_lower'],
                         color='dimgray', alpha=0.2)

    # Add a vertical line if the starting time of data2 is later than data1
    if data2[time_col].iloc[0] > data1[time_col].iloc[0]:
        plt.axvline(x=data2[time_col].iloc[0], color='black', linestyle='--')

    # Add shaded region indicating the valid period of data2
    valid_data_max = data2[target_col].max()
    valid_period = [data2[time_col].iloc[0], data2[time_col].iloc[-1]]
    plt.axvspan(valid_period[0], valid_period[1], ymin=0, ymax=valid_data_max,
                facecolor='lightyellow', alpha=0.5)

    # Add grid lines
    plt.grid(True)

    # Disable outer box
    plt.box(False)

    # Automatically adjust date label formats and positions to avoid overlap
    plt.gcf().autofmt_xdate()

    # Add legend
    plt.legend()

    # Display the plot
    plt.show()


@ParameterTypeAssert({
    'series': pd.DataFrame,
    'time_col': str,
    'target_col': str,
    'label': (None, str),
    'date_fmt': str
})
def plot_single_series(series, time_col, target_col, label=None, date_fmt='%Y-%m-%d'):
    """
    Visualize time-series data.

    Parameters
    ----------
    series : pd.DataFrame
        The dataset containing time-series data for plotting.
    time_col : str
        Column name in the dataframes representing the time information.
    target_col : str
        Column name in the dataframes representing the target variable.
    label : None or str, optional, default: None
        Label for the plot. If None, default label 'Data' will be used.
    date_fmt : str, optional, default: '%Y-%m-%d'
        Date format for the x-axis labels.

    """

    import matplotlib.pyplot as plt
    import matplotlib.dates as m_dates

    if label is None:
        label = 'Data'
    else:
        raise_if_not(TypeError, isinstance(label, str), "Label must be a string.")

    # Set x-axis as date format
    date_fmt = m_dates.DateFormatter(date_fmt)
    plt.gca().xaxis.set_major_formatter(date_fmt)

    # Plot curves
    plt.plot_date(series[time_col], series[target_col], label=label, color='black', fmt='')

    # Add grid lines
    plt.grid(True)

    # Disable outer box
    plt.box(False)

    # Automatically adjust date label formats and positions to avoid overlap
    plt.gcf().autofmt_xdate()

    # Add legend
    plt.legend()

    # Display the plot
    plt.show()
